# The MIT License (MIT)
#
# Copyright (c) 2025 FourCIPP Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Dict utils."""

from collections.abc import Callable, Iterator, Sequence
from typing import Any

import numpy as np
from loguru import logger


def compare_nested_dicts_or_lists(
    obj: Any,
    reference_obj: Any,
    allow_int_vs_float_comparison: bool = False,
    rtol: float = 1.0e-5,
    atol: float = 1.0e-8,
    equal_nan: bool = False,
    custom_compare: Callable | None = None,
) -> bool:
    """Recursively compare two nested dictionaries or lists.

    In case objects are not within the provided tolerance an `AssertionError` is raised.

    To compare custom python objects, a `custom_compare` callable can be provided which:
        - Returns nothing/`None` if the objects where not compared within `custom_compare`
        - Returns `True` if the objects are seen as equal
        - Raises AssertionError if the objects are not equal

    Args:
        obj: Object for comparison
        reference_obj: Reference object
        allow_int_vs_float_comparison: Allow a tolerance based comparison between int and
                                              float
        rtol: The relative tolerance parameter for numpy.isclose
        atol: The absolute tolerance parameter for numpy.isclose
        equal_nan: Whether to compare NaN's as equal for numpy.isclose
        custom_compare: Callable to compare objects within this nested framework

    Returns:
        True if the dictionaries are equal
    """
    # Compare non-standard python objects
    if custom_compare is not None:
        # Check if result is not None
        if result := custom_compare(obj, reference_obj) is not None:
            return result

    # Ensures the types are the same
    if not type(obj) is type(reference_obj):
        if (
            not allow_int_vs_float_comparison  # Except floats can be ints
            or not isinstance(obj, (float, int))
            or not isinstance(reference_obj, (float, int))
        ):
            raise AssertionError(
                f"Object is of type {type(obj)}, but the reference is of type {type(reference_obj)}"
            )

    # Objects are numerics
    if isinstance(obj, (float, int)):
        if not np.isclose(obj, reference_obj, rtol, atol, equal_nan):
            raise AssertionError(
                f"The numerics are not close:\n\nobj = {obj}\n\nreference_obj = {reference_obj}"
            )
        return True

    # Object are dicts
    if isinstance(obj, dict):
        # ^ is the symmetric difference operator, i.e. union of the sets without the intersection
        if missing_keys := set(obj.keys()) ^ set(reference_obj.keys()):
            raise AssertionError(
                f"The following keys could not be found in both dicts {missing_keys}:"
                f"\nobj: {obj}\n\nreference_obj:{reference_obj}"
            )
        for key in obj:
            compare_nested_dicts_or_lists(
                obj[key],
                reference_obj[key],
                allow_int_vs_float_comparison,
                rtol,
                atol,
                equal_nan,
                custom_compare,
            )
        return True

    # Objects are lists
    if isinstance(obj, list):
        if len(obj) != len(reference_obj):
            raise AssertionError(
                f"The list lengths differ (got {len(obj)} and {len(reference_obj)}).\nobj "
                f"{obj}\n\nreference_obj:{reference_obj}"
            )
        for obj_item, reference_obj_item in zip(obj, reference_obj):
            compare_nested_dicts_or_lists(
                obj_item,
                reference_obj_item,
                allow_int_vs_float_comparison,
                rtol,
                atol,
                equal_nan,
                custom_compare,
            )
        return True

    # Otherwise compare the objects directly
    if obj != reference_obj:
        raise AssertionError(
            f"The objects are not equal:\n\nobj = {obj}\n\nreference_obj = {reference_obj}"
        )

    return True


def _get_dict(
    nested_dict: dict | list, keys: Sequence, optional: bool = True
) -> Iterator[dict]:
    """Return dict entry within a nested dict by keys.

    In case a list is encountered, this function yields over every entry.

    Args:
        nested_dict: dict to iterate. Due to recursiveness, this can also be a list
        keys: List of keys to access
        optional: If the entry is part of a collection that does no exist as it is optional

    Yields:
        Desired data
    """
    # Start with the original data
    sub_data = nested_dict
    sub_keys = list(keys)

    # Get from dict
    if isinstance(nested_dict, dict):
        # Loop over all keys
        for key in keys:
            # Jump into the dict
            if isinstance(sub_data, dict):
                if key in sub_data:
                    # Jump into the entry key
                    sub_data = sub_data[key]
                    sub_keys.pop(0)
                else:
                    # Unknown key
                    if optional:
                        logger.debug(f"Entry {keys} not found and is set as optional")
                        return
                    else:
                        raise KeyError(
                            f"Key '{key}' not found in dictionary {sub_data}."
                        )
            else:
                # Jump into the sub_data with the remaining keys
                yield from _get_dict(sub_data, sub_keys)

                # Exit function afterwards
                return

    # Check the last entry type
    # dict: nothing to do
    if isinstance(sub_data, dict):
        yield sub_data
    # List: jump in an do it all over
    elif isinstance(sub_data, list):
        # Last key is a list of objects
        if not sub_keys:
            for item in sub_data:
                # Only dicts are allowed
                if isinstance(item, dict):
                    yield item
                else:
                    raise TypeError(f"Expected type dict, got {type(item)}")
        # More nested keys
        else:
            for item in sub_data:
                yield from _get_dict(item, sub_keys)
    # Unsupported type
    else:
        raise TypeError(
            f"The current data {sub_data} for type {type(sub_data)} for keys {keys} is neither a dict nor a list."
        )

    # Exit function afterwards
    return


def _split_off_last_key(
    nested_dict: dict,
    keys: Sequence,
    optional: bool = True,
    yield_dict_if_missing: bool = False,
) -> Iterator[Any]:
    """Utility to return the last key and its parent entry.

    Args:
        nested_dict: dict to iterate. Due to recursiveness, this can also be a list
        keys: List of keys to access
        optional: If the entry is part of a collection that does no exist as it is optional
        yield_dict_if_missing: Return parent entry even if the entry is not provided

    Yields:
        Parent entry
    """
    last_key = keys[-1]

    for entry in _get_dict(nested_dict, keys[:-1], optional):
        # Key has to be provided
        if last_key not in entry:
            if optional:
                logger.debug(f"Entry {keys} not found and was set to optional.")
                # Still return the dict
                if yield_dict_if_missing:
                    yield entry, last_key
                # Ignore
                else:
                    continue
            else:
                raise KeyError(f"Entry {last_key} not in {entry}")

        yield entry, last_key


def get_entry(
    nested_dict: dict,
    keys: Sequence,
    optional: bool = True,
) -> Iterator[Any]:
    """Get entry by a list of keys.

    Args:
        nested_dict: Nested data dict
        keys: List of keys to get the entry
        optional: If the entry is part of a collection that does no exist as it is optional

    Yields:
        Entry
    """
    for entry, last_key in _split_off_last_key(nested_dict, keys, optional):
        yield entry[last_key]


def remove(
    nested_dict: dict,
    keys: Sequence,
) -> None:
    """Remove entry.

    Args:
        nested_dict: Nested data dict
        keys: List of keys to the entry
        optional: If the entry is part of a collection that does no exist as it is optional
    """
    for entry, last_key in _split_off_last_key(nested_dict, keys):
        entry.pop(last_key)


def replace_value(
    nested_dict: dict,
    keys: Sequence,
    new_value: Any,
) -> None:
    """Replace value.

    Args:
        nested_dict: Nested data dict
        keys: List of keys to the entry
        new_value: New value to set
    """
    for entry, last_key in _split_off_last_key(nested_dict, keys):
        logger.debug(f"Replacing {last_key}: from {entry[last_key]} to {new_value}")
        entry[last_key] = new_value


def make_default_explicit(
    nested_dict: dict,
    keys: Sequence,
    default_value: Any,
) -> None:
    """Make default explicit, i.e. set the value in the input.

    Args:
        nested_dict: Nested data dict
        keys: List of keys to the entry
        default_value: Default value to set
    """
    for entry, last_key in _split_off_last_key(
        nested_dict, keys, yield_dict_if_missing=True
    ):
        if last_key not in entry:
            entry[last_key] = default_value


def make_default_implicit(
    nested_dict: dict,
    keys: Sequence,
    default_value: Any,
) -> None:
    """Make default implicit, i.e., removed it if set with the default value in
    the input.

    Args:
        nested_dict: Nested data dict
        keys: List of keys to the entry
        default_value: Default value to set
    """

    for entry, last_key in _split_off_last_key(nested_dict, keys):
        if entry[last_key] == default_value:
            entry.pop(last_key)


def change_default(
    nested_dict: dict,
    keys: Sequence,
    old_default: Any,
    new_default: Any,
) -> None:
    """Change default value.

    If default value is not provided the old default is set. Entries where the value equals the new default value is removed.

    Args:
        nested_dict: Nested data dict
        keys: List of keys to the entry
        old_default: Old default value to set
        new_default: New default value
    """
    for entry, last_key in _split_off_last_key(
        nested_dict, keys, yield_dict_if_missing=True
    ):
        # Optional entry is missing
        if last_key not in entry:
            entry[last_key] = old_default
        # If entry is set with the new default, remove it
        else:
            if entry[last_key] == new_default:
                entry.pop(last_key)


def rename_parameter(
    nested_dict: dict,
    keys: Sequence,
    new_name: str,
) -> None:
    """Rename parameter.

    Args:
        nested_dict: Nested data dict
        keys: List of keys to the entry
        new_name: New name of the parameter
    """

    for entry, last_key in _split_off_last_key(nested_dict, keys):
        entry[new_name] = entry.pop(last_key)


def sort_by_key_order(data: dict, key_order: list[str]) -> dict:
    """Sort a dictionary by a specific key order.

    Args:
        data: Dictionary to sort.
        key_order: List of keys in the desired order.

    Returns:
        Sorted dictionary.
    """
    if set(key_order) != set(data.keys()):
        raise ValueError("'key_order' must include all keys in the dictionary!")

    return {key: data[key] for key in key_order if key in data}


def sort_alphabetically(
    data: dict,
) -> dict:
    """Sort a dictionary alphabetically.

    Args:
        data: Dictionary to sort.

    Returns:
        Sorted dictionary.
    """
    return sort_by_key_order(
        data, sorted(data.keys(), key=lambda s: (s.lower(), 0 if s.islower() else 1))
    )



# The MIT License (MIT)
#
# Copyright (c) 2025 FourCIPP Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""4C input file handler."""

from __future__ import annotations

import copy
import difflib
import pathlib
from collections.abc import Sequence
from typing import Any, Callable

from loguru import logger

from fourcipp import CONFIG
from fourcipp.legacy_io import (
    inline_legacy_sections,
    interpret_legacy_section,
)
from fourcipp.utils.converter import Converter
from fourcipp.utils.dict_utils import (
    compare_nested_dicts_or_lists,
    sort_by_key_order,
)
from fourcipp.utils.not_set import NOT_SET, check_if_set
from fourcipp.utils.type_hinting import Path
from fourcipp.utils.validation import ValidationError, validate_using_json_schema
from fourcipp.utils.yaml_io import dump_yaml, load_yaml

# Converter for the FourCInput
CONVERTER = Converter()


class UnknownSectionException(Exception):
    """Unknown section exception."""


def is_section_known(section_name: str, known_section_names: list[str]) -> bool:
    """Returns if section in known.

    Does not apply to legacy sections.

    Args:
        section_name: Name of the section to check
        known_section_names: Names of known sections

    Returns:
        True if section is known.
    """
    return section_name in known_section_names or section_name.startswith("FUNCT")


def sort_by_section_names(data: dict) -> dict:
    """Sort a dictionary by its 4C sections.

    This sorts the dictionary in the following style:

        1. "TITLE" section
        2. Required sections (in schema order)
        3. Typed sections (alphabetically, case-insensitive)
            3.1 MATERIALS section
            3.2 'DESIGN *' sections (alphabetically, case-insensitive)
        4. FUNCT sections (numeric order)
        5. Legacy sections (alphabetically)

    Args:
        data: Dictionary to sort.
        section_names: List of all section names in the 4C style order.

    Returns:
        Dict sorted in 4C fashion
    """

    required_sections = CONFIG.fourc_json_schema["required"]
    n_sections_splitter = len(CONFIG.sections.all_sections) * 1000

    # typed sections (sorted alphabetically + case insensitive, 'DESIGN *' + 'MATERIALS' at the end)
    design_sections = [
        s for s in CONFIG.sections.typed_sections if s.startswith("DESIGN")
    ]
    remaining_typed_sections = list(
        set(CONFIG.sections.typed_sections) - set(design_sections) - set(["MATERIALS"])
    )

    typed_sections = (
        sorted(remaining_typed_sections, key=str.lower)
        + ["MATERIALS"]
        + sorted(design_sections, key=str.lower)
    )

    # function sections (sorted numerically)
    functions = sorted(
        [s for s in data.keys() if s.startswith("FUNCT") and s[5:].isdigit()],
        key=lambda s: (
            s.lower() if not s.startswith("FUNCT") else f"funct{s[5:].zfill(10)}"
        ),
    )

    def ordering_score(section: str) -> int:
        """Get ordering score, small score comes first, larger comes later.

        We offset the score by the number of sections multiplied by 1000. This way a score is guaranteed to never appear twice.

        Args:
            section: Section name to score

        Returns:
            ordering score
        """
        # Title sections
        if section == CONFIG.fourc_metadata["metadata"]["description_section_name"]:
            return 0
        # Required sections
        elif section in required_sections:
            return 1 * n_sections_splitter + required_sections.index(section)
        # Typed sections (alphabetical + case insensitive)
        elif section in typed_sections:
            return 2 * n_sections_splitter + typed_sections.index(section)
        # Function sections (numeric order)
        elif section in functions:
            return 3 * n_sections_splitter + functions.index(section)
        # Legacy sections
        elif section in CONFIG.sections.legacy_sections:
            return 4 * n_sections_splitter + CONFIG.sections.legacy_sections.index(
                section
            )
        # Unknown section
        else:
            raise KeyError(f"Unknown section {section}")

    unknown_sections = set(data.keys()) - set(CONFIG.sections.all_sections)

    # Remove functions, these are a special case
    if [section for section in unknown_sections if not section.startswith("FUNCT")]:
        raise ValueError(
            f"Sections {list(unknown_sections)} are not known in 'section_names'"
        )

    return sort_by_key_order(data, sorted(data.keys(), key=ordering_score))


class FourCInput:
    """4C inout file object."""

    # All known sections
    all_sections_names: list[str] = CONFIG.sections.all_sections

    # Legacy sections, these are not supported in the 4C JSON schema
    legacy_sections_names: list[str] = CONFIG.sections.legacy_sections

    # All sections for which the types are known, aka, non-legacy
    typed_sections_names: list[str] = CONFIG.sections.typed_sections

    type_converter: Converter = CONVERTER

    def convert_to_native_types(self) -> None:
        """Convert all sections to native Python types."""
        self._sections: dict = self.type_converter(self._sections)
        self._legacy_sections: dict = self.type_converter(self._legacy_sections)

    def __init__(
        self,
        sections: dict | None = None,
    ) -> None:
        """Initialise object.

        Args:
            sections: Sections to be added
        """
        self._sections = {}
        self._legacy_sections = {}

        if sections is not None:
            for k, v in sections.items():
                self.__setitem__(k, v)

    @classmethod
    def from_4C_yaml(
        cls, input_file_path: Path, header_only: bool = False
    ) -> FourCInput:
        """Load 4C yaml file.

        Args:
            input_file_path: Path to yaml file
            header_only: Only extract header, i.e., all sections except the legacy ones

        Returns:
            Initialised object
        """
        data = load_yaml(input_file_path)
        if header_only:
            for section in cls.legacy_sections_names:
                data.pop(section, None)
        return cls(data)

    @property
    def inlined(self) -> dict:
        """Get as dict with inlined legacy sections.

        Returns:
            dict: With all set sections in inline dat style
        """
        return self._sections | inline_legacy_sections(
            self._legacy_sections.copy(), self.legacy_sections_names
        )

    def __repr__(self) -> str:
        """Representation string.

        Returns:
            str: Representation string
        """
        string = "\n4C Input file"
        string += "\n with sections\n  - "
        string += "\n  - ".join(self.get_section_names()) + "\n"
        return string

    def __str__(self) -> str:
        """To string method,

        Returns:
            str: Object description.
        """
        string = "\n4C Input file"
        string += "\n with sections\n  - "
        string += "\n  - ".join(self.get_section_names()) + "\n"
        return string

    def __setitem__(
        self,
        key: str,
        value: Any,
    ) -> None:
        """Set section.

        Args:
            key: Section name
            value: Section entry
        """
        value = self.type_converter(value)
        # Warn if complete section is overwritten
        if key in self.sections:
            logger.warning(f"Section {key} was overwritten.")
        # Nice sections
        if is_section_known(key, self.typed_sections_names):
            self._sections[key] = value
        # Legacy sections
        elif key in self.legacy_sections_names:
            # Is a list needs to be interpreted to dict
            if isinstance(value, list):
                if not any([isinstance(v, dict) for v in value]):
                    logger.debug(f"Interpreting section {key}")
                    self._legacy_sections[key] = interpret_legacy_section(
                        key, value, self.legacy_sections_names
                    )
                else:
                    # Sections are in dict form
                    self._legacy_sections[key] = value
            elif isinstance(value, dict):
                self._legacy_sections[key] = value
            else:
                raise TypeError(f"Section {key} is not a list or dict.")

        else:
            # Fancy error message
            raise UnknownSectionException(
                f"Unknown section '{key}'. Did you mean "
                f"'{difflib.get_close_matches(key.upper(), self.all_sections_names, n=1, cutoff=0.3)[0]}'?"
                " Call FourCInputFile.known_sections for a complete list."
            )

    def __getitem__(self, key: str) -> Any:
        """Get section.

        Args:
            key: Section name

        Returns:
            Section value
        """
        # Nice sections
        if is_section_known(key, self.typed_sections_names):
            return self._sections[key]
        # Legacy sections
        elif key in self._legacy_sections:
            return self._legacy_sections[key]
        else:
            sections = "\n - ".join(self.get_section_names())
            raise UnknownSectionException(
                f"Section '{key}' not set. Did out mean '{difflib.get_close_matches(key.upper(), self.all_sections_names, n=1, cutoff=0.3)[0]}'? The set sections are:\n - {sections}"
            )

    def pop(self, key: str, default_value: Any = NOT_SET) -> Any:
        """Pop entry.

        Args:
            key: Section name
            default_value: Default value if section is not set

        Returns:
            Desired section or default value
        """
        # Section is set
        if key in self._sections:
            return self._sections.pop(key)
        elif key in self._legacy_sections:
            return self._legacy_sections.pop(key)
        # Section is not set
        else:
            # Known section
            if key in self.all_sections_names:
                # Default value was provided
                if check_if_set(default_value):
                    return default_value
                # Default value was not provided
                else:
                    raise UnknownSectionException(
                        f"Section '{key}' not set. Did out mean '{difflib.get_close_matches(key.upper(), self.get_section_names(), n=1, cutoff=0.3)[0]}'?"
                    )
            # Unknown section
            else:
                raise UnknownSectionException(
                    f"Unknown section '{key}'. Did you mean "
                    f"'{difflib.get_close_matches(key.upper(), self.all_sections_names, n=1, cutoff=0.3)[0]}'?"
                    " Call FourCInputFile.known_sections for a complete list."
                )

    def combine_sections(self, other: dict | FourCInput) -> None:
        """Combine input files together.

        Note: Every section can only be defined in self or in other.

        Args:
            other: Sections to be combined
        """
        other_sections_names: Any = None

        if isinstance(other, dict):
            other_sections_names = other.keys()
        elif isinstance(other, FourCInput):
            other_sections_names = other.get_section_names()
        else:
            raise TypeError(
                f"Cannot combine sections between {type(self)} and {type(other)}."
            )

        # Sections that can be found in both
        if doubled_defined_sections := set(self.get_section_names()) & set(
            other_sections_names  # type: ignore
        ):
            raise ValueError(
                f"Section(s) {', '.join(list(doubled_defined_sections))} are defined in both {type(self).__name__} objects. In order to join the {type(self).__name__} objects remove the section(s) in one of them."
            )

        self.overwrite_sections(other)

    def overwrite_sections(self, other: dict | FourCInput) -> None:
        """Overwrite sections from dict or FourCInput.

        This function always overwrites complete sections. Combining parameters within
        sections has to be done manually.


        Args:
            other: Sections to be updated
        """
        if isinstance(other, (dict, FourCInput)):
            for key, value in other.items():
                self[key] = value
        else:
            raise TypeError(f"Cannot overwrite sections from {type(other)}.")

    def apply_user_defaults(
        self, default_path: Path | None = CONFIG.user_defaults_path
    ) -> None:
        """Combines two Inputs by overwriting current values by a file
        containing user defaults.

        This function checks whether values exist in both objects and overwrites the current by the other.
        At this time only top level section parameters of simple types (int, float, str, bool, None) are supported.

        Args:
            default_path: String containing the path to the YAML file with user defaults
        """
        if default_path is None:
            raise ValueError(f"User defaults path is not set in the config: {CONFIG}")
        logger.info(f"Applying user defaults from '{default_path}''.")
        user_defaults_path = pathlib.Path(default_path)
        default_input = FourCInput.from_4C_yaml(user_defaults_path, header_only=True)
        default_sections = default_input.sections

        for section_key in default_sections:
            if not isinstance(default_sections[section_key], dict):
                raise TypeError(f"Section {section_key} does not contain a dict.")
            if section_key in self.sections:
                # only check sections that are contained in the current object and in the default_sections
                for parameter_key in default_sections[section_key]:
                    if parameter_key not in self[section_key]:
                        self[section_key][parameter_key] = default_sections[
                            section_key
                        ][parameter_key]
                        logger.debug(
                            "Setting user default value {default_sections[section_key]} to parameter {parameter_key} in section {section_key}"
                        )
                        continue
                    if (
                        not isinstance(
                            self[section_key][parameter_key], (int, float, str, bool)
                        )
                        and self[section_key][parameter_key] is not None
                    ):
                        print(
                            "At this time, you should only use the default values for parameters in top level section!"
                        )
                        raise TypeError(
                            f"The value for parameter {parameter_key} in section {section_key} is not a primitive."
                        )
            else:
                # take the section content from default if it is not in the current object
                self[section_key] = default_sections[section_key]
                logger.debug(f"Adding user default section {section_key} to input file")

    @property
    def sections(self) -> dict:
        """All the set sections.

        Returns:
            dict: Set sections
        """
        return self._sections | self._legacy_sections

    def get_section_names(self) -> list:
        """Get set section names.

        Returns:
            list: Sorted section names
        """
        return sorted(list(self._legacy_sections) + list(self._sections))

    def items(self) -> Any:
        """Get items.

        Similar to items method of python dicts.

        Returns:
            dict_items: Dict items
        """
        return (self.sections).items()

    def __contains__(self, item: str) -> bool:
        """Contains function.

        Allows to use the `in` operator.

        Args:
            item: Section name to check if it is set

        Returns:
            True if section is set
        """
        return item in (list(self._legacy_sections) + list(self._sections))

    def __add__(self, other: FourCInput) -> FourCInput:
        """Add two input file objects together.

        In contrast to `join` a copy is created.

        Args:
            other: Input file object to join.

        Returns:
            Joined input file
        """
        copied_object = self.copy()
        copied_object.combine_sections(other)
        return copied_object

    def copy(self) -> FourCInput:
        """Copy itself.

        Returns:
            FourCInputFile: Copy of current object
        """
        return copy.deepcopy(self)

    def load_includes(self) -> None:
        """Load data from the includes section."""
        if includes := self.pop("INCLUDES", None):
            for partial_file in includes:
                logger.debug(f"Gather data from {partial_file}")
                self.combine_sections(self.from_4C_yaml(partial_file))

    def dump(
        self,
        input_file_path: Path,
        validate: bool = False,
        validate_sections_only: bool = False,
        convert_to_native_types: bool = True,
        sort_function: Callable[[dict], dict] | None = sort_by_section_names,
        use_fourcipp_yaml_style: bool = True,
    ) -> None:
        """Dump object to yaml.

        Args:
            input_file_path: Path to dump the data to
            validate: Validate input data before dumping
            validate_sections_only: Validate each section independently.
                Requiredness of the sections themselves is ignored.
            convert_to_native_types: Convert all sections to native Python types
            sort_function: Function to sort the sections.
            use_fourcipp_yaml_style: If FourCIPP yaml style is to be used
        """

        if validate or validate_sections_only:
            self.validate(
                sections_only=validate_sections_only,
                convert_to_native_types=convert_to_native_types,
            )
            # if conversion already happened in validation do not convert again
            if convert_to_native_types:
                convert_to_native_types = False

        if convert_to_native_types:
            self.convert_to_native_types()

        dump_yaml(self.inlined, input_file_path, sort_function, use_fourcipp_yaml_style)

    def validate(
        self,
        json_schema: dict = CONFIG.fourc_json_schema,
        sections_only: bool = False,
        convert_to_native_types: bool = True,
    ) -> bool:
        """Validate input file.

        Args:
            json_schema: Schema to check the data
            sections_only: Validate each section independently.
                Requiredness of the sections themselves is ignored.
            convert_to_native_types: Convert all sections to native Python types
        """
        validation_schema = json_schema

        # Remove the requiredness of the sections
        if sections_only:
            validation_schema = json_schema.copy()
            validation_schema.pop("required")

        if convert_to_native_types:
            self.convert_to_native_types()

        # Validate sections using schema
        validate_using_json_schema(self._sections, validation_schema)

        # Legacy sections are only checked if they are of type string
        for section_name, section in inline_legacy_sections(
            self._legacy_sections.copy(), self.legacy_sections_names
        ).items():
            for i, k in enumerate(section):
                if not isinstance(k, str):
                    raise ValidationError(
                        f"Could not validate the legacy section {section_name}, since entry {i}:\n{k} is not a string"
                    )

        return True

    def split(self, section_names: Sequence) -> tuple[FourCInput, FourCInput]:
        """Split input into two using sections names.

        Args:
            section_names: List of sections to split

        Returns:
            root and split input objects
        """
        root_input = self.copy()
        spiltted_input = FourCInput()

        for section in section_names:
            spiltted_input[section] = root_input.pop(section)

        return root_input, spiltted_input

    def dump_with_includes(
        self,
        section_names: Sequence,
        root_input_file_path: Path,
        split_input_file_path: Path,
        invert_sections: bool = False,
        sort_sections: bool = False,
        validate: bool = False,
    ) -> None:
        """Dump input and split using the includes function.

        Args:
            section_names: List of sections to split
            root_input_file_path: Directory with the INCLUDES section
            split_input_file_path: Remaining sections
            invert_sections: Switch sections in root and split file
            sort_sections: Sort the sections alphabetically
            validate: Validate input data before dumping
        """
        # Split the inout
        first_input, second_input = self.split(section_names)

        # Select where the input should be
        if not invert_sections:
            input_with_includes = first_input
            split_input = second_input
        else:
            split_input = first_input
            input_with_includes = second_input

        # Add includes sections if missing
        if "INCLUDES" not in input_with_includes:
            input_with_includes["INCLUDES"] = []

        # Append the path to the second file
        input_with_includes["INCLUDES"].append(str(split_input_file_path))

        # Dump files
        input_with_includes.dump(root_input_file_path, sort_sections, validate)
        split_input.dump(split_input_file_path, sort_sections, validate)

    def __eq__(self, other: object) -> bool:
        """Define equal operator.

        This comparison is strict, if tolerances are desired use `compare`.

        Args:
            other: Other input to check
        """
        if not isinstance(other, type(self)):
            raise TypeError(f"Can not compare types {type(self)} and {type(other)}")

        return self.sections == other.sections

    def compare(
        self,
        other: FourCInput,
        allow_int_as_float: bool = False,
        rtol: float = 1.0e-5,
        atol: float = 1.0e-8,
        equal_nan: bool = False,
        raise_exception: bool = False,
    ) -> bool:
        """Compare inputs with tolerances.

        Args:
            other: Input to compare
            allow_int_as_float: Allow the use of ints instead of floats
            rtol: The relative tolerance parameter for numpy.isclose
            atol: The absolute tolerance parameter for numpy.isclose
            equal_nan: Whether to compare NaN's as equal for numpy.isclose
            raise_exception: If true raise exception

            Returns:
            True if within tolerance
        """
        try:
            return compare_nested_dicts_or_lists(
                other.sections, self.sections, allow_int_as_float, rtol, atol, equal_nan
            )
        except AssertionError as exception:
            if raise_exception:
                raise AssertionError(
                    "Inputs are not equal or within tolerances"
                ) from exception

            return False

    def extract_header(self) -> FourCInput:
        """Extract the header sections, i.e., all non-legacy sections.

        Returns:
            FourCInput: Input with only the non-legacy sections
        """
        return FourCInput(sections=self._sections)



# The MIT License (MIT)
#
# Copyright (c) 2025 FourCIPP Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Converter to convert custom types to native Python types."""

from __future__ import annotations

from collections.abc import Callable
from typing import Any

import numpy as np


class Converter:
    """Converter class to convert custom types to native Python types."""

    def __init__(self) -> None:
        self._custom_converters: dict = {}

    def register_type(
        self,
        custom_type: type,
        converter_function: Callable[[Converter, Any], Any],
    ) -> Converter:
        """Register a custom type and its converter function.

        The first argument of the converter_function is a converter object.
        This allows you to pass self down to your converter function to
        recursively call it and use custom types you already registered.
        The second argument is the object to be converted.

        Args:
            custom_type: Custom class to register
            converter_function: Converter function
        """
        self._custom_converters[custom_type] = converter_function
        return self

    def register_types(self, types_dict: dict) -> Converter:
        """Register multiple custom types and their converter functions.

        Args:
            types_dict: Dictionary with custom types as keys and
            converter functions as values
        """
        self._custom_converters.update(types_dict)
        return self

    def register_numpy_types(self) -> Converter:
        """Register NumPy types and their converter functions."""

        def convert_ndarray(converter: Converter, obj: np.ndarray) -> list[Any]:
            """Convert a NumPy ndarray to a list.

            Args:
                converter: Converter object
                obj: NumPy ndarray to convert
            """
            return converter(obj.tolist())

        def convert_generic(
            converter: Converter, obj: np.generic
        ) -> int | float | bool | str:
            """Convert a NumPy generic type to a native Python type.

            Args:
                converter: Converter object
                obj: NumPy generic type to convert
            """
            return obj.item()

        self.register_type(np.generic, convert_generic)
        self.register_type(np.ndarray, convert_ndarray)
        return self

    def __call__(self, obj: Any) -> Any:
        """Convert the object to a native Python type.

        Args:
            obj: Object to convert
        """
        # If no custom converters are present, no need to do a conversion
        if not self._custom_converters:
            return obj

        # Look if object is present in the custom converters
        for custom_type in self._custom_converters:
            if isinstance(obj, custom_type):
                # First match will be used.
                # Keep in mind for inheritance you have think about child classes!
                # Make sure if you have parent and child classes registered separately, to first register the child classes!
                return self._custom_converters[custom_type](self, obj)

        # Lets convert
        match obj:
            # Convert the nested types
            case list():
                return [self(entry) for entry in obj]
            case set():
                return (self(entry) for entry in obj)
            case dict():
                return {k: self(v) for k, v in obj.items()}

            # Nothing to do here, since these are the accepted types
            case int() | float() | bool() | str() | None:
                return obj

            # Type was not registered and is not one of the standards one
            case _:
                raise TypeError(
                    f"Object {obj} of type {type(obj)} can not be converted"
                )

    def __str__(self) -> str:
        """String representation of the Converter class."""
        string = "Converter with custom object (objects will be converted from top to bottom):"
        for k, v in self._custom_converters.items():
            string += f"\n\t{k}\t: {v}"

        return string



# The MIT License (MIT)
#
# Copyright (c) 2025 FourCIPP Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Metadata objects from the 4C input metadata."""

from __future__ import annotations

import abc
import re
from collections.abc import Generator, Sequence
from typing import Callable, Literal, Protocol, TypeAlias, TypeVar

from fourcipp.utils.not_set import NotSet, check_if_set
from fourcipp.utils.type_hinting import NotSetAlias

PRIMITIVES_PYTHON_TYPES = float | bool | int | str

ValidatorAlias: TypeAlias = Callable[[object], bool] | None

NotSetString: NotSetAlias[str] = NotSet(str)

T = TypeVar("T", int, float)
G = TypeVar("G", contravariant=True)


class InputSpec:
    """Input spec base class."""

    def __init__(
        self,
        spec_type: str,
        name: NotSetAlias[str] = NotSetString,
        description: NotSetAlias[str] = NotSetString,
        required: bool = False,
        noneable: bool = False,
        validator: ValidatorAlias = None,
    ) -> None:
        """Initialise InputSpec.

        Args:
            spec_type (str): Type of input spec
            name: Name
            description: Description
            required: True if parameter is required
            noneable: True if parameter can be None
            validator: Validator callable
        """
        self.spec_type = spec_type
        self.name = name
        self.description = description
        self.required = required
        self.noneable = noneable
        self.validator = validator

    @classmethod
    @abc.abstractmethod
    def from_4C_metadata(cls, data_dict: dict) -> InputSpec:
        """Create InputSpec from 4C metadata file.

        Args:
            data_dict: Data from 4C metadata file

        Returns:
            InputSpec
        """


class All_Of:
    def __init__(
        self,
        specs: Sequence[Spec],
        description: NotSetAlias[str] = NotSetString,
    ) -> None:
        """Initialise All_Of.

        Args:
            specs: Specs of the All_Of
            description: Description
        """
        self.specs: list[InputSpec | One_Of] = All_Of.condense(specs)
        self.description = description

    @staticmethod
    def condense(specs: Sequence[Spec]) -> list[InputSpec | One_Of]:
        """Condense All_Of.

        All_Ofs of All_Ofs are joined, aka flattened or condensed.

        Args:
            specs: Specs to condense

        Returns:
            Condensed specs
        """
        new_specs: list[InputSpec | One_Of] = []
        one_ofs: list[One_Of] = []

        for spec in specs:
            match spec:
                case InputSpec():
                    new_specs.append(spec)
                case All_Of():
                    # Check if it is an One_Of
                    if spec.is_one_of():
                        one_ofs.append(spec.specs[0])  # type: ignore
                    else:
                        new_specs.extend(All_Of.condense(spec.specs))
                case One_Of():
                    one_ofs.append(spec)
                case _:
                    raise TypeError(f"Invalid spec {spec}")

        if len(one_ofs) > 1:
            oo = "\n".join([str(oo) for oo in one_ofs])
            raise ValueError(f"More than one_of is not allowed {oo}")

        if one_ofs:
            new_specs = [one_ofs[0].add_specs(new_specs)]

        return new_specs

    def add_specs(self, input_specs: Sequence[Spec]) -> All_Of:
        """Add specs to the All_Of.

        Returns:
            All_Of object.
        """
        self.specs = All_Of.condense(self.specs + list(input_specs))
        return self

    def is_one_of(self) -> bool:
        """Check if All_Of only hold a One_Of.

        In essence this means that the All_Of is simply a One_Of.

        Returns:
            True if is a One_Of in disguise
        """
        return len(self.specs) == 1 and isinstance(self.specs[0], One_Of)

    @classmethod
    def from_4C_metadata(cls, data_dict: dict) -> All_Of:
        """Create All_Of from 4C metadata file.

        Args:
            data_dict: Data from 4C metadata file

        Returns:
            All_Of
        """
        data_dict.pop("type", None)
        specs = data_dict.pop("specs", [])
        specs = [metadata_from_dict(spec) for spec in specs]
        return cls(specs=specs, **data_dict)

    def __str__(self) -> str:  # pragma: no cover
        """String method."""
        return "All_of(" + ", ".join([str(spec) for spec in self.specs]) + ")"

    def __repr__(self) -> str:  # pragma: no cover
        """Representation method."""
        return "All_of(" + ", ".join([str(spec) for spec in self.specs]) + ")"

    def __len__(self) -> int:
        """Len of the All_Of.

        Returns:
            number of specs
        """
        return len(self.specs)

    def __iter__(self) -> Generator[One_Of | InputSpec]:
        """Iterate All_Of.

        Yields:
            All_Of specs
        """
        yield from self.specs


class One_Of:
    def __init__(
        self,
        specs: Sequence[Spec],
        description: NotSetAlias[str] = NotSetString,
    ) -> None:
        """Initialise One_Of.

        Args:
            specs: Specs of the One_Of, i.e., each distinct branch
            description: Description
        """
        self.specs: list[All_Of] = One_Of.condense(specs)
        self.description = description

    @staticmethod
    def condense(specs: Sequence[Spec]) -> list[All_Of]:
        """Condense One_Of.

        One_Ofs of One_Ofs are combined to a single One_Of.

        Args:
            specs: Specs of the One_Of

        Returns:
            condesed specs
        """
        new_specs: list[All_Of] = []

        for spec in specs:
            match spec:
                case InputSpec():
                    new_specs.append(All_Of([spec]))
                case All_Of():
                    if spec.is_one_of():
                        # One_of in One_of
                        new_specs.extend(One_Of.condense(spec.specs[0].specs))  # type: ignore
                    else:
                        new_specs.append(spec)
                case One_Of():
                    # One_of in one_of
                    new_specs.extend(One_Of.condense(spec.specs))
                case _:
                    raise TypeError(f"Not supported type {type(spec)} for {spec}")

        return new_specs

    def add_specs(self, input_specs: Sequence[Spec]) -> One_Of:
        """Add specs to the One_Of.

        Returns:
            One_Of object.
        """
        for i in range(len(self.specs)):
            self.specs[i].add_specs(input_specs)
        self.specs = One_Of.condense(self.specs)
        return self

    @classmethod
    def from_4C_metadata(cls, data_dict: dict) -> One_Of:
        """Create One_Of from 4C metadata file.

        Args:
            data_dict: Data from 4C metadata file

        Returns:
            One_Of
        """
        data_dict.pop("type", None)
        specs = [metadata_from_dict(spec) for spec in data_dict.pop("specs")]
        return cls(specs=specs, **data_dict)

    def __str__(self) -> str:  # pragma: no cover
        """String method."""
        return "One_of(" + ", ".join([str(spec) for spec in self.specs]) + ")"

    def __repr__(self) -> str:  # pragma: no cover
        """Representation method."""
        return "One_of(" + ", ".join([str(spec) for spec in self.specs]) + ")"

    def __len__(self) -> int:
        """Len of the One_Of.

        Returns:
            number of One_Of branches
        """
        return len(self.specs)

    def __iter__(self) -> Generator[All_Of]:
        """Iterate One_of.

        Yields:
            One_Of branches
        """
        yield from self.specs


class Primitive(InputSpec):
    PRIMITIVE_TYPES: list[str] = ["double", "bool", "int", "string", "path"]
    PRIMITIVE_TYPES_TO_PYTHON: dict[str, type | Callable] = {
        "double": float,
        "bool": bool,
        "int": int,
        "string": str,
        "path": str,
    }

    def __init__(
        self,
        spec_type: Literal["double", "int", "bool", "string", "path"],
        name: NotSetAlias[
            str
        ] = NotSetString,  # Name is only optional with vectors, maps, tuple
        description: NotSetAlias[str] = NotSetString,
        required: bool = True,
        noneable: bool = False,
        validator: ValidatorAlias = None,
        default: NotSetAlias[PRIMITIVES_PYTHON_TYPES] = NotSet("default"),
    ) -> None:
        """Initialise primitive.

        Args:
            spec_type: Type of primitive
            name: Name
            description: Description
            required: True if parameter is required
            noneable: True if parameter can be None
            validator: Validator callable
            default: Default value of the spec
        """
        if spec_type not in Primitive.PRIMITIVE_TYPES:
            raise TypeError(
                f"Spec type {spec_type} is not supported. Supported ones are: {Primitive.PRIMITIVE_TYPES}"
            )
        super().__init__(spec_type, name, description, required, noneable, validator)
        if check_if_set(default) and default is not None and False:
            if not isinstance(default, Primitive.PRIMITIVE_TYPES_TO_PYTHON[spec_type]):
                raise TypeError(
                    f"Default {default} for the parameter of type '{spec_type}' is not Python type {Primitive.PRIMITIVE_TYPES_TO_PYTHON[spec_type]}"
                )
        self.default = default

    @classmethod
    def from_4C_metadata(cls, data_dict: dict) -> Primitive:
        """Create primitive from 4C metadata file.

        Args:
            data_dict: Data from 4C metadata file

        Returns:
            primitive
        """
        spec_type = data_dict.pop("type")
        if "validator" in data_dict:
            data_dict["validator"] = validator_from_dict(data_dict["validator"])
        return cls(spec_type=spec_type, **data_dict)

    def __str__(self) -> str:  # pragma: no cover
        """String method."""
        return f"{self.name}"

    def __repr__(self) -> str:  # pragma: no cover
        """Representation method."""
        return f"{self.name}"


class Enum(InputSpec):
    def __init__(
        self,
        choices: Sequence[str],
        name: NotSetAlias[str] = NotSetString,
        description: NotSetAlias[str] = NotSetString,
        required: bool = True,
        noneable: bool = False,
        validator: ValidatorAlias = None,
        default: NotSetAlias[str] = NotSet("default"),
    ) -> None:
        """Initialise enum.

        Args:
            choices: Enum choices
            name: Name
            description: Description
            required: True if parameter is required
            noneable: True if parameter can be None
            validator: Validator callable
            default: Default value
        """
        super().__init__("enum", name, description, required, noneable, validator)
        if check_if_set(default):
            if default not in choices:
                raise ValueError(
                    f"Default choice {default} is not in the valid enum choices {choices}"
                )
        self.choices = choices
        self.default = default

    @classmethod
    def from_4C_metadata(cls, data_dict: dict) -> Enum:
        """Create enum from 4C metadata file.

        Args:
            data_dict: Data from 4C metadata file

        Returns:
            enum
        """
        data_dict.pop("type", None)
        choices = [c["name"] for c in data_dict.pop("choices")]
        # TODO add description from choices
        if "validator" in data_dict:
            data_dict["validator"] = validator_from_dict(data_dict["validator"])
        return cls(choices=choices, **data_dict)

    def __str__(self) -> str:  # pragma: no cover
        """String method."""
        return f"{self.name}"

    def __repr__(self) -> str:  # pragma: no cover
        """Representation method."""
        return f"{self.name}"


class Vector(InputSpec):
    def __init__(
        self,
        value_type: NATIVE_CPP_ALIAS,
        size: int | None = None,
        name: NotSetAlias[str] = NotSetString,
        description: NotSetAlias[str] = NotSetString,
        required: bool = True,
        noneable: bool = False,
        validator: ValidatorAlias = None,
        default: NotSetAlias[Sequence[NATIVE_CPP_ALIAS]] = NotSet("default"),
    ) -> None:
        """Initialise vector.

        Args:
            value_type: Type of vector elements
            size: Vector dimension
            name: Name
            description: Description
            required: True if parameter is required
            noneable: True if parameter can be None
            validator: Validator callable
            default: Default value
        """
        super().__init__("vector", name, description, required, noneable, validator)
        if not isinstance(value_type, NATIVE_CPP_TYPES):
            raise TypeError(
                f"Value type {value_type} has to be of type: {NATIVE_CPP_TYPES}"
            )
        self.value_type = value_type
        self.size = size
        self.default = default

    @classmethod
    def from_4C_metadata(cls, data_dict: dict) -> Vector:
        """Create vector from 4C metadata file.

        Args:
            data_dict: Data from 4C metadata file

        Returns:
            vector
        """
        data_dict.pop("type", None)
        value_type: NATIVE_CPP_ALIAS = metadata_from_dict(data_dict.pop("value_type"))  # type: ignore
        if "validator" in data_dict:
            data_dict["validator"] = validator_from_dict(data_dict["validator"])
        return cls(value_type=value_type, **data_dict)

    def __str__(self) -> str:  # pragma: no cover
        """String method."""
        return f"{self.name}"

    def __repr__(self) -> str:  # pragma: no cover
        """Representation method."""
        return f"{self.name}"

    def __iter__(self) -> Generator[NATIVE_CPP_ALIAS]:
        """Iterate vector.

        Yields:
            value type of the vector
        """
        yield self.value_type


class Map(InputSpec):
    def __init__(
        self,
        value_type: NATIVE_CPP_ALIAS,
        size: int | None = None,
        name: NotSetAlias[str] = NotSetString,
        description: NotSetAlias[str] = NotSetString,
        required: bool = True,
        noneable: bool = False,
        validator: ValidatorAlias = None,
        default: NotSetAlias[dict[str, NATIVE_CPP_ALIAS]] = NotSet("default"),
    ) -> None:
        """Initialise map.

        Args:
            value_type: Type of map elements
            size: Number of entries in the map
            name: Name
            description: Description
            required: True if parameter is required
            noneable: True if parameter can be None
            validator: Validator callable
            default: Default value
        """
        if not isinstance(value_type, NATIVE_CPP_TYPES):
            raise TypeError(
                f"Value type {value_type} has to be of type: {NATIVE_CPP_TYPES}"
            )
        super().__init__("map", name, description, required, noneable, validator)
        self.value_type: NATIVE_CPP_ALIAS = value_type
        self.size = size
        self.default = default

    @classmethod
    def from_4C_metadata(cls, data_dict: dict) -> Map:
        """Create map from 4C metadata file.

        Args:
            data_dict: Data from 4C metadata file

        Returns:
            map
        """
        data_dict.pop("type", None)
        value_type: NATIVE_CPP_ALIAS = metadata_from_dict(data_dict.pop("value_type"))  # type: ignore
        if "validator" in data_dict:
            data_dict["validator"] = validator_from_dict(data_dict["validator"])
        return cls(value_type=value_type, **data_dict)

    def __str__(self) -> str:  # pragma: no cover
        """String method."""
        return f"{self.name}"

    def __repr__(self) -> str:  # pragma: no cover
        """Representation method."""
        return f"{self.name}"

    def __iter__(self) -> Generator[NATIVE_CPP_ALIAS]:
        """Iterate map.

        Yields:
            value type of the map
        """
        yield self.value_type


class Tuple(InputSpec):
    def __init__(
        self,
        value_types: Sequence[NATIVE_CPP_ALIAS],
        size: int,
        name: NotSetAlias[str] = NotSetString,
        description: NotSetAlias[str] = NotSetString,
        required: bool = True,
        noneable: bool = False,
        validator: ValidatorAlias = None,
        default: NotSetAlias[Sequence[NATIVE_CPP_ALIAS]] = NotSet("default"),
    ) -> None:
        """Initialise tuple.

        Args:
            value_types: Types of the tuple element
            size: Number of tuple entries
            name: Name
            description: Description
            required: True if parameter is required
            noneable: True if parameter can be None
            validator: Validator callable
            default: Default value
        """
        super().__init__("tuple", name, description, required, noneable, validator)
        for i, value_type in enumerate(value_types):
            if not isinstance(value_type, NATIVE_CPP_TYPES):
                raise TypeError(
                    f"Value type {value_type} (parameter number {i + 1} in {value_types}) has to be of type: {NATIVE_CPP_TYPES}"
                )
        self.value_types = value_types
        self.size = size
        if self.size != len(self.value_types):
            raise ValueError(
                f"Tuple size {self.size} does not match the number of value types "
                f"{len(self.value_types)}"
            )
        self.default = default

    @classmethod
    def from_4C_metadata(cls, data_dict: dict) -> Tuple:
        """Create tuple from 4C metadata file.

        Args:
            data_dict: Data from 4C metadata file

        Returns:
            tuple
        """
        data_dict.pop("type", None)
        value_types = [
            metadata_from_dict(value_type)
            for value_type in data_dict.pop("value_types")
        ]
        if "validator" in data_dict:
            data_dict["validator"] = validator_from_dict(data_dict["validator"])
        return cls(value_types=value_types, **data_dict)  # type: ignore

    def __str__(self) -> str:  # pragma: no cover
        """String method."""
        return f"{self.name}"

    def __repr__(self) -> str:  # pragma: no cover
        """Representation method."""
        return f"{self.name}"

    def __iter__(self) -> Generator[NATIVE_CPP_ALIAS]:
        """Iterate tuple.

        Yields:
            value types of the tuple
        """
        yield from self.value_types


class Selection(InputSpec):
    def __init__(
        self,
        name: NotSetAlias[str],
        choices: dict[str, Spec],
        description: NotSetAlias[str] = NotSetString,
        required: bool = True,
        noneable: bool = False,
        defaultable: bool = False,
        validator: ValidatorAlias = None,
    ) -> None:
        """Initialise selection.

        Args:
            name: Name
            choices: Choices of the selection
            description: Description
            required: True if parameter is required
            noneable: True if parameter can be None
            defaultable: Can have a default
            validator: Validator callable
        """
        super().__init__("selection", name, description, required, noneable, validator)
        self.choices: dict[str, All_Of] = {k: All_Of([v]) for k, v in choices.items()}
        self.defautable = defaultable

    @classmethod
    def from_4C_metadata(cls, data_dict: dict) -> Selection:
        """Create selection from 4C metadata file.

        Args:
            data_dict: Data from 4C metadata file

        Returns:
            selection
        """
        data_dict.pop("type", None)

        choices = {}
        for choice in data_dict.pop("choices"):
            spec = choice.pop("spec")
            choices[choice.pop("name")] = metadata_from_dict(spec)
        if "validator" in data_dict:
            data_dict["validator"] = validator_from_dict(data_dict["validator"])
        return cls(choices=choices, **data_dict)

    def __iter__(self) -> Generator[Spec]:
        """Iterate selection.

        Yields:
            choices of the selection
        """
        yield from self.choices.values()

    def __len__(self) -> int:
        """Len of the selection.

        Returns:
            number of choices
        """
        return len(self.choices)


class Group(InputSpec):
    def __init__(
        self,
        name: NotSetAlias[str],
        spec: Spec,
        description: NotSetAlias[str] = NotSetString,
        required: bool = True,
        noneable: bool = False,
        defaultable: bool = False,
        validator: ValidatorAlias = None,
    ) -> None:
        """Initialise group.

        Args:
            name: Name
            spec: Spec of the group
            description: Description
            required: True if parameter is required
            noneable: True if parameter can be None
            defaultable: Can have a default
            validator: Validator callable
        """
        super().__init__("group", name, description, required, noneable, validator)
        self.spec: All_Of = All_Of([spec])
        self.defautable = defaultable

    @classmethod
    def from_4C_metadata(cls, data_dict: dict) -> Group:
        """Create group from 4C metadata file.

        Args:
            data_dict: Data from 4C metadata file

        Returns:
            group
        """
        data_dict.pop("type", None)
        specs = [metadata_from_dict(spec) for spec in data_dict.pop("specs", [])]
        if "validator" in data_dict:
            data_dict["validator"] = validator_from_dict(data_dict["validator"])
        return cls(spec=All_Of(specs), **data_dict)

    def __iter__(self) -> Generator[All_Of]:
        """Iterate group.

        Yields:
            group spec
        """
        yield self.spec

    def __str__(self) -> str:  # pragma: no cover
        """String method."""
        return f"{self.name} Group({self.spec})"

    def __repr__(self) -> str:  # pragma: no cover
        """Representation method."""
        return f"{self.name} Group({self.spec})"


class List(InputSpec):
    def __init__(
        self,
        name: NotSetAlias[str],
        spec: Spec,
        size: int | None = None,
        description: NotSetAlias[str] = NotSetString,
        required: bool = True,
        noneable: bool = False,
        defaultable: bool = False,
        validator: ValidatorAlias = None,
    ) -> None:
        """Initialise list.

        Args:
            name: Name
            spec: Spec of the list entries
            size: Number of list entries
            description: Description
            required: True if parameter is required
            noneable: True if parameter can be None
            defaultable: Can have a default
            validator: Validator callable
        """
        super().__init__("list", name, description, required, noneable, validator)
        self.spec: All_Of = All_Of([spec])
        self.size = size
        self.defautable = defaultable

    @classmethod
    def from_4C_metadata(cls, data_dict: dict) -> List:
        """Create list from 4C metadata file.

        Args:
            data_dict: Data from 4C metadata file

        Returns:
            list
        """
        data_dict.pop("type", None)
        spec = metadata_from_dict(data_dict.pop("spec"))
        if "validator" in data_dict:
            data_dict["validator"] = validator_from_dict(data_dict["validator"])
        return cls(spec=All_Of([spec]), **data_dict)

    def __iter__(self) -> Generator[All_Of]:
        """Iterate list.

        Yields:
            spec
        """
        yield self.spec

    def __str__(self) -> str:  # pragma: no cover
        """String method."""
        return f"{self.name} List({self.spec})"

    def __repr__(self) -> str:  # pragma: no cover
        """Representation method."""
        return f"{self.name} List({self.spec})"


Spec: TypeAlias = InputSpec | All_Of | One_Of
NATIVE_CPP_ALIAS: TypeAlias = Primitive | Enum | Vector | Map | Tuple
NATIVE_CPP_TYPES = (Primitive, Enum, Vector, Map, Tuple)

_METADATA_TYPE_TO_CLASS: dict[str, type[Spec]] = {
    i.__name__.lower(): i
    for i in [Enum, Vector, Map, Tuple, Selection, Group, List, All_Of, One_Of]
} | {i: Primitive for i in Primitive.PRIMITIVE_TYPES}


def metadata_from_dict(data_dict: dict) -> Spec:
    """Create metadata object from 4C metadata dict.

    Args:
        data_dict: Data from 4C metadata file

    Returns:
        metadata object
    """
    entry_type: str = data_dict.get("type")  # type: ignore

    metadata_class: type[Spec] | None = _METADATA_TYPE_TO_CLASS.get(entry_type, None)

    if metadata_class is not None:
        return metadata_class.from_4C_metadata(data_dict)

    raise TypeError(
        f"Unknown type {entry_type} for {data_dict}, known ones are {list(_METADATA_TYPE_TO_CLASS.keys())}"
    )


class Validator(Protocol[G]):
    """Validator protocol."""

    def _validate(self, entry: G) -> bool:
        """Validate object.

        Args:
            entry: Object to validate

        Returns:
            True if valid
        """
        pass

    def __call__(self, entry: G) -> bool:
        """Call method."""
        return self._validate(entry)


class RangeValidator(Validator[T]):
    def __init__(
        self,
        minimum: T,
        maximum: T,
        minimum_exclusive: bool = False,
        maximum_exclusive: bool = False,
    ) -> None:
        """Range validator.

        Args:
            entry: Entry value
            minimum: Minimum value
            maximum: Maximum value
            minimum_exclusive: Minimum is exclusive
            maximum_exclusive: Maximum is exclusive
        """
        self.minimum: T = minimum
        self.maximum: T = maximum
        self.minimum_exclusive = minimum_exclusive
        self.maximum_exclusive = maximum_exclusive

    def _validate(self, entry: T) -> bool:
        """Validate if float or in is in range.

        Args:
            entry: Entry value
            minimum: Minimum value
            maximum: Maximum value
            minimum_exclusive: Minimum is exclusive
            maximum_exclusive: Maximum is exclusive

        Returns:
            True if in range
        """
        if self.minimum_exclusive:
            if entry <= self.minimum:
                return False
        else:
            if entry < self.minimum:
                return False

        if self.maximum_exclusive:
            if entry >= self.maximum:
                return False
        else:
            if entry > self.maximum:
                return False

        return True

    @classmethod
    def from_4C_metadata_dict(cls, data_dict: dict) -> RangeValidator:
        """Create validator from 4C metadata dict.

        Args:
            data_dict: Data from 4C metadata file

        Returns:
            range validator
        """
        return cls(**data_dict)


class AllEmementsValidator(Validator[Sequence]):
    def __init__(self, element_validator: Callable[[Sequence], bool]) -> None:
        """All elements validator.

        Validates each entry of a sequence with the `element_validator`.

        Args:
            element_validator: Validator for each entry
        """
        self.element_validator = element_validator

    def _validate(self, entry: Sequence) -> bool:
        """Validate sequence.

        Args:
            entry: Sequence to validate

        Returns:
            True if all entries are valid
        """
        return all([self.element_validator(e) for e in entry])

    @classmethod
    def from_4C_metadata_dict(cls, data_dict: dict) -> AllEmementsValidator:
        """Create validator from 4C metadata dict.

        Args:
            data_dict: Data from 4C metadata file

        Returns:
            all elements validator
        """
        element_validator = validator_from_dict(data_dict)
        return cls(element_validator)


class PatternValidator(Validator):
    def __init__(self, pattern: str):
        """Regex pattern validator.

        Args:
            pattern: Pattern to match
        """
        self.pattern = re.compile(pattern)

    def _validate(self, entry: str) -> bool:
        """Validate pattern.

        Args:
            entry: Entry to validate

        Returns:
            True if pattern matches
        """
        return bool(self.pattern.match(entry))

    @classmethod
    def from_4C_metadata_dict(cls, data_dict: dict) -> PatternValidator:
        """Create validator from 4C metadata dict.

        Args:
            data_dict: Data from 4C metadata file

        Returns:
            validator
        """
        return cls(**data_dict)


def validator_from_dict(data_dict: dict) -> Validator:
    """Create validator from dict.

    Args:
        data_dict: Data from 4C metadata file

    Returns:
        validator
    """
    if len(data_dict) == 0:
        raise ValueError(f"Can not construct value error from {data_dict}.")
    validator_name = list(data_dict.keys())[0]
    validator_data = data_dict[validator_name]
    match validator_name:
        case "range":
            return RangeValidator.from_4C_metadata_dict(validator_data)
        case "all_elements":
            return AllEmementsValidator.from_4C_metadata_dict(validator_data)
        case "pattern":
            return PatternValidator.from_4C_metadata_dict(validator_data)
        case _:
            raise KeyError(f"Unknown validator {data_dict}")



# The MIT License (MIT)
#
# Copyright (c) 2025 FourCIPP Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Testing framework infrastructure."""

from _pytest.config.argparsing import Parser


def pytest_addoption(parser: Parser) -> None:
    """Add custom command line options to pytest.

    Args:
        parser: Pytest parser
    """
    parser.addoption(
        "--performance-tests",
        action="store_true",
        default=False,
        help="Execute standard and performance tests.",
    )


def pytest_collection_modifyitems(config, items):
    """Filter tests based on their markers and provided command line options.

    Currently configured options:
        `pytest`: Execute standard tests with no markers
        `pytest --performance-tests`: Execute standard tests and tests with the `performance` marker

    Args:
        config: Pytest config
        items: Pytest list of tests
    """

    selected_tests = []

    # loop over all collected tests
    for item in items:
        # Get all set markers for the current test (ignore standard markers)
        markers = [
            marker.name
            for marker in item.iter_markers()
            if marker.name not in {"parametrize", "skipif", "skip", "xfail"}
        ]

        # Add the performance test if the option is set and the marker is present
        if config.getoption("--performance-tests") and "performance" in markers:
            selected_tests.append(item)

        if not markers:
            selected_tests.append(item)

    deselected_tests = list(set(items) - set(selected_tests))

    items[:] = selected_tests
    config.hook.pytest_deselected(items=deselected_tests)



# The MIT License (MIT)
#
# Copyright (c) 2025 FourCIPP Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""CLI utils."""

import argparse
import pathlib
import sys

from loguru import logger

from fourcipp import CONFIG
from fourcipp.fourc_input import FourCInput
from fourcipp.utils.configuration import (
    change_profile,
    show_config,
)
from fourcipp.utils.type_hinting import Path
from fourcipp.utils.yaml_io import dump_yaml, load_yaml


def modify_input_with_defaults(
    input_path: Path, overwrite: bool
) -> None:  # pragma: no cover
    """Apply user defaults to an input file located at input_path.

    Args:
         input_path: Input filename to apply user defaults to.
         overwrite: Whether to overwrite the existing input file.
                         By default, a new file with suffix '_mod.4C.yaml' is created.
    """
    output_appendix = "_mod"

    input_path = pathlib.Path(input_path)
    if not input_path.is_file():
        raise FileNotFoundError(f"Input file '{input_path}' does not exist.")
    input_data = FourCInput.from_4C_yaml(input_path)
    user_defaults_string = CONFIG.user_defaults_path
    input_data.apply_user_defaults(user_defaults_string)
    if overwrite:
        output_filename = input_path
    else:
        names = input_path.name.split(".")
        names[0] += output_appendix
        output_filename = input_path.parent / ".".join(names)
    input_data.dump(output_filename)
    logger.info(f"Input file incl. user defaults is now '{output_filename}'.")


def format_file(
    input_file: str, sort_sections: bool = False
) -> None:  # pragma: no cover
    """Formatting file.

    Args:
        input_file: File to format
        sort_sections: Sort sections
    """
    if sort_sections:
        # Requires reading the config
        fourc_input = FourCInput.from_4C_yaml(input_file)
        fourc_input.dump(input_file, use_fourcipp_yaml_style=True)
    else:
        # No config required, is purely a style question
        dump_yaml(load_yaml(input_file), input_file, use_fourcipp_yaml_style=True)


def main() -> None:
    """Main CLI interface."""
    # Set up the logger
    logger.enable("fourcipp")
    logger.remove()
    logger.add(sys.stdout, format="{message}")

    # The FourCIPP CLI is build upon argparse and subparsers. The latter ones are use to interface
    # mutual exclusive commands. If you add a new command add a new subparser and add the CLI
    # parameters as you would normally with argparse. Finally add the new command to the pattern
    # matching. More details can be found here:
    # https://docs.python.org/3/library/argparse/html#sub-commands

    main_parser = argparse.ArgumentParser(prog="FourCIPP")
    subparsers = main_parser.add_subparsers(dest="command")

    # Config parser
    subparsers.add_parser("show-config", help="Show the FourCIPP config.")

    # Switch config parser
    switch_config_profile_parser = subparsers.add_parser(
        "switch-config-profile", help="Switch user config profile."
    )
    switch_config_profile_parser.add_argument(
        "profile",
        help=f"FourCIPP config profile name.",
        type=str,
    )

    # Apply user defaults parser
    apply_user_defaults_parser = subparsers.add_parser(
        "apply-user-defaults",
        help="Apply user defaults from the file given in the user defaults path.",
    )

    apply_user_defaults_parser.add_argument(
        "-o",
        "--overwrite",
        action="store_true",
        help=f"Overwrite existing input file.",
    )

    apply_user_defaults_parser.add_argument(
        "input-file",
        help=f"4C input file.",
        type=str,
    )

    # Format parser
    format_parser = subparsers.add_parser(
        "format",
        help="Format the file in fourcipp style. This sorts the sections and uses the flow styles from FourCIPP",
    )

    format_parser.add_argument(
        "input-file",
        help=f"4C input file.",
        type=str,
    )

    format_parser.add_argument(
        "--sort-sections",
        action="store_true",
        help=f"Overwrite existing input file.",
    )
    # Replace "-" with "_" for variable names
    kwargs: dict = {}
    for key, value in vars(main_parser.parse_args(sys.argv[1:])).items():
        kwargs[key.replace("-", "_")] = value
    command = kwargs.pop("command")

    # Select the desired command
    match command:
        case "show-config":
            show_config()
        case "switch-config-profile":
            change_profile(**kwargs)
        case "apply-user-defaults":
            input_path = pathlib.Path(kwargs.pop("input_file"))
            overwrite = kwargs.pop("overwrite")
            modify_input_with_defaults(input_path, overwrite)
        case "format":
            format_file(**kwargs)



# The MIT License (MIT)
#
# Copyright (c) 2025 FourCIPP Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Test metadata utils."""

import copy

import pytest

from fourcipp import CONFIG
from fourcipp.utils.metadata import (
    All_Of,
    AllEmementsValidator,
    Enum,
    Group,
    List,
    Map,
    One_Of,
    PatternValidator,
    Primitive,
    RangeValidator,
    Selection,
    Tuple,
    Vector,
)
from fourcipp.utils.not_set import check_if_set


@pytest.mark.parametrize("primitive_type", Primitive.PRIMITIVE_TYPES)
def test_primitive(primitive_type):
    """Assert if primitives are set correctly."""
    primitive = Primitive(spec_type=primitive_type, name="primitive_" + primitive_type)
    assert primitive.spec_type == primitive_type
    assert primitive.name == "primitive_" + primitive_type
    assert not check_if_set(primitive.default)
    assert not check_if_set(primitive.description)
    assert primitive.validator is None
    assert primitive.required
    assert not primitive.noneable


@pytest.mark.parametrize("primitive_type", Primitive.PRIMITIVE_TYPES)
def test_primitive_from_4C_metadata(primitive_type):
    """Test primitives from metadata."""
    metadata_dict = {
        "type": primitive_type,
        "name": "primitive_" + primitive_type,
        "description": "this is the description",
        "required": False,
    }
    primitive = Primitive.from_4C_metadata(metadata_dict)
    assert primitive.spec_type == primitive_type
    assert primitive.name == "primitive_" + primitive_type
    assert primitive.description == "this is the description"
    assert not check_if_set(primitive.default)
    assert primitive.validator is None
    assert not primitive.required
    assert not primitive.noneable


def test_primitive_invalid_type():
    """Test if invalid type can be set."""
    with pytest.raises(TypeError, match="Spec type this_is_not_valid"):
        Primitive(spec_type="this_is_not_valid")


@pytest.mark.skip(reason="Validation is not fully implemented yet.")
@pytest.mark.parametrize("primitive_type", Primitive.PRIMITIVE_TYPES)
def test_primitive_invalid_default_type(primitive_type):
    """Test if invalid default type."""
    invalid_default = "invalid_default"
    if primitive_type in ["string", "path"]:
        invalid_default = 42
    with pytest.raises(TypeError, match="Default"):
        Primitive(spec_type=primitive_type, default=invalid_default)


def test_enum():
    """Test enum."""
    choices = ["choice_1", "choice_2", "choice_3"]
    enum = Enum(name="enum_name", choices=choices)
    assert enum.name == "enum_name"
    assert enum.choices == choices
    assert not check_if_set(enum.default)
    assert not check_if_set(enum.description)
    assert enum.validator is None
    assert enum.required
    assert not enum.noneable


def test_enum_from_4C_metadata():
    """Test enum from dict."""
    metadata_dict = {
        "name": "enum_name",
        "type": "enum",
        "description": "enum description.",
        "required": False,
        "default": "choice_1",
        "choices": [{"name": "choice_1"}, {"name": "choice_2"}, {"name": "choice_3"}],
    }
    enum = Enum.from_4C_metadata(metadata_dict)
    assert enum.name == "enum_name"
    assert enum.choices == ["choice_1", "choice_2", "choice_3"]
    assert enum.default == "choice_1"
    assert enum.description == "enum description."
    assert enum.validator is None
    assert not enum.required
    assert not enum.noneable


def test_enum_invalid_default_choice():
    """Test invalid default choice."""
    choices = ["choice_1", "choice_2", "choice_3"]
    with pytest.raises(
        ValueError,
        match="Default choice invalid_choice is not in the valid enum choices",
    ):
        Enum(name="enum_name", choices=choices, default="invalid_choice")


@pytest.mark.parametrize("metadata_class", (Vector, Map))
@pytest.mark.parametrize("primitive_type", Primitive.PRIMITIVE_TYPES)
def test_vector_and_maps_primitives(metadata_class, primitive_type):
    """Test vector and map from primitives."""
    metadata_object = metadata_class(Primitive(primitive_type))
    assert isinstance(metadata_object.value_type, Primitive)
    assert not check_if_set(metadata_object.default)
    assert not check_if_set(metadata_object.description)
    assert metadata_object.validator is None
    assert metadata_object.required
    assert not metadata_object.noneable


@pytest.mark.parametrize(
    "metadata_type,metadata_class", (("vector", Vector), ("map", Map))
)
def test_vector_and_maps_from_4C_metadata(metadata_type, metadata_class):
    """Test vector and map from 4C metadata."""
    metadata_dict = {
        "name": metadata_type + "_parameter",
        "type": metadata_type,
        "size": 3,
        "value_type": {"type": "double"},
        "required": True,
    }

    metadata_object = metadata_class.from_4C_metadata(metadata_dict)

    assert isinstance(metadata_object, metadata_class)
    assert isinstance(metadata_object.value_type, Primitive)
    assert metadata_object.name == metadata_type + "_parameter"
    assert not check_if_set(metadata_object.default)
    assert not check_if_set(metadata_object.description)
    assert metadata_object.validator is None
    assert metadata_object.required
    assert not metadata_object.noneable


@pytest.mark.parametrize("metadata_class", (Vector, Map))
def test_vector_and_maps_enum(metadata_class):
    """Test vector and map from enum."""
    metadata_object = metadata_class(Enum(choices=["a", "b"]), size=3)
    assert isinstance(metadata_object.value_type, Enum)
    assert not check_if_set(metadata_object.default)
    assert not check_if_set(metadata_object.description)
    assert metadata_object.validator is None
    assert metadata_object.required
    assert not metadata_object.noneable
    assert metadata_object.size == 3


@pytest.mark.parametrize("metadata_class", (Vector, Map))
def test_nested_vector(metadata_class):
    """Test nested vector and map."""
    metadata_object = metadata_class(
        metadata_class(metadata_class(Primitive("bool")), description="second inner")
    )
    assert isinstance(metadata_object.value_type.value_type.value_type, Primitive)
    assert metadata_object.value_type.description == "second inner"
    assert not check_if_set(metadata_object.default)
    assert not check_if_set(metadata_object.description)
    assert metadata_object.validator is None
    assert metadata_object.required
    assert not metadata_object.noneable


@pytest.mark.parametrize("metadata_class", (Vector, Map))
def test_invalid_value_type_vector(metadata_class):
    """Test invalid value type in vector and maps."""
    with pytest.raises(TypeError, match="Value type invalid_object has to be of type"):
        metadata_class(metadata_class("invalid_object"))


def test_tuple():
    """Test tuple."""
    tuple_object = Tuple((Enum(["choice_1", "choice_2"]), Primitive("int")), size=2)
    assert isinstance(tuple_object.value_types[0], Enum)
    assert isinstance(tuple_object.value_types[1], Primitive)
    assert not check_if_set(tuple_object.default)
    assert not check_if_set(tuple_object.description)
    assert tuple_object.validator is None
    assert tuple_object.required
    assert not tuple_object.noneable


def test_tuple_failure_size():
    """Test failure of the tuple initialization."""
    with pytest.raises(ValueError):
        Tuple((Enum(["choice_1", "choice_2"]), Primitive("int")), size=3)


def test_tuple_invalid_value_types():
    """Test invalid value types in tuples."""
    with pytest.raises(TypeError, match="Value type not a valid type"):
        Tuple((Enum(["choice_1", "choice_2"]), "not a valid type"), size=2)


def test_tuple_from_4C_metadata():
    """Test tuple from 4C metadata."""
    metadata_dict = {
        "value_types": [
            {
                "name": "enum_name",
                "type": "enum",
                "description": "enum description.",
                "required": False,
                "default": "choice_1",
                "choices": [
                    {"name": "choice_1"},
                    {"name": "choice_2"},
                    {"name": "choice_3"},
                ],
            },
            {
                "name": "vector_name",
                "type": "vector",
                "size": 3,
                "value_type": {"type": "double"},
                "required": True,
            },
        ],
        "size": 2,
    }
    tuple_object = Tuple.from_4C_metadata(metadata_dict)
    assert isinstance(tuple_object.value_types[0], Enum)
    assert isinstance(tuple_object.value_types[1], Vector)
    assert not check_if_set(tuple_object.default)
    assert not check_if_set(tuple_object.description)
    assert tuple_object.validator is None
    assert tuple_object.required
    assert not tuple_object.noneable


def test_all_of_from_simple_types():
    """Test All_of from simple types."""
    all_of = All_Of(specs=[Enum(["choice_1", "choice_2"]), Primitive("int")])
    assert isinstance(all_of.specs[0], Enum)
    assert isinstance(all_of.specs[1], Primitive)
    assert not check_if_set(all_of.description)


def test_nested_all_of():
    """Test All_of from simple types and All_of."""
    all_of = All_Of(
        specs=[
            All_Of(
                specs=[
                    Enum(["choice_1", "choice_2"]),
                    Primitive("int", "a"),
                    All_Of(
                        specs=[
                            All_Of(
                                specs=[Primitive("double", "b")],
                            )
                        ]
                    ),
                ]
            ),
            All_Of(specs=[Primitive("bool", "c")]),
        ]
    )
    assert isinstance(all_of.specs[0], Enum)
    assert all_of.specs[1].name == "a"
    assert all_of.specs[2].name == "b"
    assert all_of.specs[3].name == "c"


def test_nested_all_of_with_one_of():
    """Test All_of from simple types, All_of and One_of."""
    all_of = All_Of(
        specs=[
            All_Of(
                specs=[
                    Primitive("int", "a"),
                    All_Of(
                        specs=[
                            All_Of(
                                specs=[
                                    Primitive("double", "b"),
                                    One_Of(
                                        [
                                            Primitive("int", "d"),
                                            Primitive("double", "e"),
                                        ]
                                    ),
                                ],
                            )
                        ]
                    ),
                ]
            ),
            All_Of(specs=[Primitive("bool", "c")]),
        ]
    )

    assert len(all_of) == 1
    assert isinstance(all_of.specs[0], One_Of)
    assert len(all_of.specs[0]) == 2
    assert isinstance(all_of.specs[0].specs[0], All_Of)
    assert isinstance(all_of.specs[0].specs[1], All_Of)
    assert all_of.specs[0].specs[0].specs[0].name == "d"
    assert all_of.specs[0].specs[0].specs[1].name == "b"
    assert all_of.specs[0].specs[0].specs[2].name == "a"
    assert all_of.specs[0].specs[0].specs[3].name == "c"
    assert all_of.specs[0].specs[1].specs[0].name == "e"
    assert all_of.specs[0].specs[1].specs[1].name == "b"
    assert all_of.specs[0].specs[1].specs[2].name == "a"
    assert all_of.specs[0].specs[1].specs[3].name == "c"


def test_add_specs_all_of_from_simple_types():
    """Test add simple input specs."""
    all_of = All_Of(specs=[Enum(["choice_1", "choice_2"]), Primitive("int", "a")])
    all_of.add_specs([Primitive("bool", "b")])

    assert isinstance(all_of.specs[0], Enum)
    assert all_of.specs[1].name == "a"
    assert all_of.specs[2].name == "b"
    assert len(all_of.specs) == 3


def test_add_specs_all_of_nested():
    """Test add nested specs with All_Ofs to All_of."""
    all_of = All_Of(specs=[Enum(["choice_1", "choice_2"]), Primitive("int", "a")])
    all_of.add_specs(
        [
            Primitive("bool", "b"),
            All_Of([Primitive("double", "c"), All_Of([Primitive("int", "d")])]),
        ]
    )

    assert isinstance(all_of.specs[0], Enum)
    assert all_of.specs[1].name == "a"
    assert all_of.specs[2].name == "b"
    assert all_of.specs[3].name == "c"
    assert all_of.specs[4].name == "d"
    assert len(all_of.specs) == 5


def test_all_of_from_4C_metadata():
    """Test All_Of from 4C metadata."""
    metadata_dict = {
        "type": "all_of",
        "specs": [
            {
                "type": "bool",
                "name": "a",
                "description": "this is the description",
                "required": False,
            },
            {
                "type": "all_of",
                "specs": [
                    {
                        "type": "int",
                        "name": "b",
                        "description": "this is the description",
                        "required": False,
                    }
                ],
            },
        ],
    }
    obj = All_Of.from_4C_metadata(metadata_dict)
    assert len(obj) == 2
    assert obj.specs[0].name == "a"
    assert obj.specs[1].name == "b"


def test_one_of_from_4C_metadata():
    """Test one of from 4C metadata."""
    metadata_dict = {
        "type": "one_of",
        "specs": [
            {
                "type": "bool",
                "name": "a",
                "description": "this is the description",
                "required": False,
            },
            {
                "type": "all_of",
                "specs": [
                    {
                        "type": "int",
                        "name": "b",
                        "description": "this is the description",
                        "required": False,
                    }
                ],
            },
        ],
    }
    obj = One_Of.from_4C_metadata(metadata_dict)
    assert len(obj) == 2
    assert len(obj.specs[0]) == 1
    assert len(obj.specs[1]) == 1
    assert obj.specs[0].specs[0].name == "a"
    assert obj.specs[1].specs[0].name == "b"


def test_one_of_from_simple_types():
    """Test one_of from simple types."""
    one_of = One_Of(specs=[Enum(["choice_1", "choice_2"], "a"), Primitive("int", "b")])

    assert not check_if_set(one_of.description)
    assert len(one_of.specs) == 2
    assert isinstance(one_of.specs[0], All_Of)
    assert isinstance(one_of.specs[1], All_Of)
    assert len(one_of.specs[0].specs) == 1
    assert len(one_of.specs[1].specs) == 1
    assert one_of.specs[0].specs[0].name == "a"
    assert one_of.specs[1].specs[0].name == "b"


def test_one_of_from_simple_types_and_all_of():
    """Test one_of from simple types and All_of."""
    one_of = One_Of(
        specs=[
            Enum(["choice_1", "choice_2"], "a"),
            All_Of([Primitive("int", "b"), Primitive("bool", "c")]),
        ]
    )

    assert len(one_of.specs) == 2
    assert isinstance(one_of.specs[0], All_Of)
    assert isinstance(one_of.specs[1], All_Of)
    assert len(one_of.specs[0].specs) == 1
    assert len(one_of.specs[1].specs) == 2
    assert one_of.specs[0].specs[0].name == "a"
    assert one_of.specs[1].specs[0].name == "b"
    assert one_of.specs[1].specs[1].name == "c"


def test_one_of_from_simple_types_and_one_of():
    """Test one_of from simple types and One_ofs."""
    one_of = One_Of(
        specs=[
            Enum(["choice_1", "choice_2"], "a"),
            One_Of([Primitive("int", "b"), All_Of([Primitive("bool", "c")])]),
        ]
    )

    assert len(one_of.specs) == 3
    assert isinstance(one_of.specs[0], All_Of)
    assert isinstance(one_of.specs[1], All_Of)
    assert isinstance(one_of.specs[2], All_Of)
    assert len(one_of.specs[0].specs) == 1
    assert len(one_of.specs[1].specs) == 1
    assert len(one_of.specs[2].specs) == 1
    assert one_of.specs[0].specs[0].name == "a"
    assert one_of.specs[1].specs[0].name == "b"
    assert one_of.specs[2].specs[0].name == "c"


def test_add_specs_one_of():
    """Add specs to One_of."""
    one_of = One_Of(specs=[Enum(["choice_1", "choice_2"], "a"), Primitive("int", "b")])
    one_of.add_specs([Primitive("double", "c"), Primitive("bool", "d")])

    assert len(one_of.specs) == 2
    assert isinstance(one_of.specs[0], All_Of)
    assert isinstance(one_of.specs[1], All_Of)
    assert len(one_of.specs[0].specs) == 3
    assert len(one_of.specs[1].specs) == 3
    assert one_of.specs[0].specs[0].name == "a"
    assert one_of.specs[0].specs[1].name == "c"
    assert one_of.specs[0].specs[2].name == "d"
    assert one_of.specs[1].specs[0].name == "b"
    assert one_of.specs[1].specs[1].name == "c"
    assert one_of.specs[1].specs[2].name == "d"


def test_selection():
    """Test selection."""
    selection = Selection(
        "selection",
        {
            "a": Primitive("double", "a"),
            "b": All_Of([Primitive("int", "b"), Primitive("bool", "c")]),
        },
    )
    assert selection.name == "selection"
    assert len(selection) == 2
    assert isinstance(selection.choices["a"], All_Of)
    assert isinstance(selection.choices["b"], All_Of)
    assert len(selection.choices["a"]) == 1
    assert selection.choices["a"].specs[0].name == "a"
    assert len(selection.choices["b"]) == 2
    assert selection.choices["b"].specs[0].name == "b"
    assert selection.choices["b"].specs[1].name == "c"


def test_selection_from_4C_metadata():
    """Test selection from 4C metadata."""
    metadata_dict = {
        "type": "selection",
        "name": "a_selection",
        "choices": [
            {"name": "choice_a", "spec": {"type": "int", "name": "a"}},
            {
                "name": "choice_b",
                "spec": {"type": "all_of", "specs": [{"type": "double", "name": "b"}]},
            },
        ],
    }
    selection = Selection.from_4C_metadata(metadata_dict)
    assert selection.name == "a_selection"
    assert len(selection) == 2
    assert isinstance(selection.choices["choice_a"], All_Of)
    assert isinstance(selection.choices["choice_b"], All_Of)
    assert len(selection.choices["choice_a"]) == 1
    assert selection.choices["choice_a"].specs[0].name == "a"
    assert len(selection.choices["choice_b"]) == 1
    assert selection.choices["choice_b"].specs[0].name == "b"


@pytest.mark.parametrize("metadata_class", (Group, List))
@pytest.mark.parametrize(
    "spec",
    [
        Primitive("int", "a"),
        All_Of([Primitive("bool", "b"), Primitive("double", "c")]),
    ],
)
def test_group_and_list(metadata_class, spec):
    """Test group and list."""
    metadata = metadata_class(metadata_class.__name__.lower(), spec)
    spec = All_Of([spec])
    assert isinstance(metadata.spec, All_Of)
    for i in range(len(metadata.spec.specs)):
        assert metadata.spec.specs[i].name == spec.specs[i].name


@pytest.mark.parametrize("metadata_class", (Group, List))
@pytest.mark.parametrize(
    "spec",
    (
        {"specs": [{"type": "int", "name": "a"}]},
        {
            "specs": [
                {
                    "type": "all_of",
                    "specs": [{"type": "int", "name": "a"}],
                }
            ]
        },
    ),
)
def test_group_and_list_from_4C_metadata(metadata_class, spec):
    """Test group and list from 4C metadata."""
    spec = copy.deepcopy(spec)
    spec["type"] = metadata_class.__name__.lower()
    spec["name"] = metadata_class.__name__.lower()
    # Difference between group and list
    if metadata_class.__name__ == "List":
        spec["spec"] = spec.pop("specs")[0]
    metadata = metadata_class.from_4C_metadata(spec)
    assert isinstance(metadata.spec, All_Of)
    assert len(metadata.spec) == 1
    assert metadata.spec.specs[0].name == "a"


def test_list_size():
    """Test list size."""
    list_metadata = List("a_list", Primitive("int", "a"), size=5)
    assert list_metadata.size == 5


def test_read_in_metadata_from_config():
    """Sanity check if metadata file can be read in."""
    sections = CONFIG.fourc_metadata["sections"]
    obj = All_Of.from_4C_metadata(sections)
    assert len(obj) > 10


@pytest.mark.parametrize("data_type", (int, float))
@pytest.mark.parametrize(
    "value,minimum_exclusive,maximum_exclusive,expected",
    [
        (2, True, True, True),
        (7, True, True, False),
        (2, False, False, True),
        (7, False, False, False),
        (1, True, True, False),
        (1, False, True, True),
        (5, True, True, False),
        (5, True, False, True),
    ],
)
def test_range_validator(
    data_type, value, minimum_exclusive, maximum_exclusive, expected
):
    """Test range validator."""
    validator = RangeValidator(
        minimum=data_type(1),
        minimum_exclusive=minimum_exclusive,
        maximum=data_type(5),
        maximum_exclusive=maximum_exclusive,
    )
    assert validator(data_type(value)) == expected


@pytest.mark.parametrize("value,expected", [("test this", True), ("abc", False)])
def test_pattern_validator(value, expected):
    """Test pattern validator."""
    validator = PatternValidator("test *")
    assert validator(value) == expected


@pytest.mark.parametrize("value,expected", [([2, 3, 4], True), ([4, 7, 2], False)])
def test_all_elements(value, expected):
    """Test all elements validator."""
    validator = AllEmementsValidator(RangeValidator(1, 4))
    assert validator(value) == expected



# The MIT License (MIT)
#
# Copyright (c) 2025 FourCIPP Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Validation utils."""

from __future__ import annotations

import sys
from collections.abc import Iterable, Sequence

import jsonschema_rs

from fourcipp.utils.yaml_io import dict_to_yaml_string

MAX_INT = 2_147_483_647  # C++ value
MAX_FLOAT = sys.float_info.max


class ValidationError(Exception):
    """FourCIPP validation error."""

    @staticmethod
    def path_indexer(path: Sequence[str | int]) -> str:
        """Create a path representation to walk the dict."""
        path_for_data = ""
        for p in path:
            if isinstance(p, str):
                p = '"' + p + '"'
            path_for_data += "[" + str(p) + "]"
        return path_for_data

    @staticmethod
    def indent(text: str, n_spaces: int = 4) -> str:
        """Indent the text."""
        indent_with_newline = "\n" + " " * n_spaces
        return indent_with_newline + text.replace("\n", indent_with_newline)

    @classmethod
    def from_schema_validation_errors(
        cls, errors: Iterable[jsonschema_rs.ValidationError]
    ) -> ValidationError:
        """Create error from multiple errors.

        Args:
            errors: Errors to raise

        Returns:
            New error for this case
        """
        message = "\nValidation failed, due to the following parameters:"

        for error in errors:
            message += "\n\n- Parameter in " + cls.path_indexer(error.instance_path)
            message += cls.indent(cls.indent(dict_to_yaml_string(error.instance), 4))
            message += cls.indent("Error: " + error.message, 2)

        return cls(message)

    @classmethod
    def from_overflow_errors(
        cls, object_paths_with_errors: Iterable[tuple[list[str | int], int | float]]
    ) -> ValidationError:
        """Create error from multiple errors.

        Args:
            errors: Errors to raise

        Returns:
            New error for this case
        """
        message = "\nValidation failed, due to the following parameters:"

        for path, obj in object_paths_with_errors:
            message += "\n\n- Parameter in " + cls.path_indexer(path)
            message += cls.indent(cls.indent(str(obj), 4))

            if isinstance(obj, int):
                error = f"Maximum int value {MAX_INT} exceeded"
            else:
                error = f"Maximum float value {MAX_FLOAT} exceeded"
            message += cls.indent("Error: " + error, 2)

        return cls(message)


def find_keys_exceeding_max_value(
    obj: object, path_for_data: list[str | int] | None = None
) -> Iterable[tuple[list[str | int], int | float]]:
    """Find entries exceeding max values.

    Args:
        obj (object): Object to check
        path_for_data: Path to the data

    Yields:
        path for each case where this problem emerges
    """
    if path_for_data is None:
        path_for_data = []

    if isinstance(obj, dict):
        for key, value in obj.items():
            yield from find_keys_exceeding_max_value(value, path_for_data + [key])
    elif isinstance(obj, list):
        for index, item in enumerate(obj):
            yield from find_keys_exceeding_max_value(item, path_for_data + [index])
    elif isinstance(obj, int) and abs(obj) > MAX_INT:
        yield path_for_data, obj
    elif isinstance(obj, float) and abs(obj) > MAX_FLOAT:
        yield path_for_data, obj


def validate_using_json_schema(data: dict, json_schema: dict) -> bool:
    """Validate data using a JSON schema.

    Args:
        data: Data to validate
        json_schema: Schema for validation

    Returns:
        True if successful
    """
    validator = jsonschema_rs.validator_for(json_schema)
    try:
        validator.validate(data)
    except jsonschema_rs.ValidationError as exception:
        # Validation failed, look for all errors
        raise ValidationError.from_schema_validation_errors(
            validator.iter_errors(data)
        ) from exception
    except ValueError as exception:
        if str(exception).endswith("too big to convert"):
            raise ValidationError.from_overflow_errors(
                find_keys_exceeding_max_value(data)
            ) from exception
        raise
    return True



# The MIT License (MIT)
#
# Copyright (c) 2025 FourCIPP Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Test Converter class."""

import numpy as np
import pytest

from fourcipp.utils.converter import Converter
from fourcipp.utils.dict_utils import compare_nested_dicts_or_lists


# create a fixture for the converter
@pytest.fixture
def converter():
    """Fixture for the Converter class."""
    return Converter()


def test_basic_python_types(converter):
    """Test identity conversion for basic Python types."""
    for value in [123, 12.5, "text", True]:
        assert converter(value) == value


def test_list_and_dict_conversion(converter):
    """Test identity conversion of nested lists and dictionaries."""
    input_data = {"a": [1, 2, {"b": "text"}], "c": True}
    assert converter(input_data) == input_data


def test_nested_numpy_structures(converter):
    """Test nested structures with NumPy types."""
    converter.register_numpy_types()

    obj = {
        "arr": np.array([1, 2, 3]),
        "nested": {"scalar": np.int64(42), "float": np.float32(3.1)},
    }

    expected = {
        "arr": [1, 2, 3],
        "nested": {
            "scalar": 42,
            "float": 3.1,
        },
    }

    assert compare_nested_dicts_or_lists(converter(obj), expected)


def test_register_custom_type(converter):
    """Test custom type registration and conversion."""

    class Custom:
        """Custom class for testing."""

        def __init__(self, value):
            self.value = value

    class Custom2:
        """Another custom class for testing."""

        def __init__(self, value):
            self.value = value

    def convert_custom(converter, obj):
        """Custom conversion function for the Custom class."""
        return {"custom_value": converter(obj.value)}

    converter.register_type(Custom, convert_custom)
    converter.register_type(Custom2, convert_custom)

    custom_obj = Custom2([Custom("ab"), Custom(3)])
    assert converter(custom_obj) == {
        "custom_value": [{"custom_value": "ab"}, {"custom_value": 3}]
    }


def test_not_convertible_type(converter):
    """Test conversion of a non-convertible type."""

    # register a type to enable conversion
    converter.register_numpy_types()

    class UnknownType:
        """Unknown type for testing."""

        pass

    with pytest.raises(TypeError, match=r"can not be converted"):
        converter(UnknownType())



# The MIT License (MIT)
#
# Copyright (c) 2025 FourCIPP Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Test yaml io utils."""

import pytest

from fourcipp.utils.dict_utils import sort_alphabetically
from fourcipp.utils.yaml_io import dict_to_yaml_string, dump_yaml, load_yaml


def test_dump_not_sorted(tmp_path):
    """Test if key order is preserved."""
    data = {"c": 1, "b": 2, "a": 3}
    sorted_file_path = tmp_path / "sorted.yaml"
    dump_yaml(data, path_to_yaml_file=sorted_file_path)
    reloaded_data = load_yaml(sorted_file_path)
    assert reloaded_data == data


def test_dump_sorted_alphabetically(tmp_path):
    """Test if key order is sorted."""
    data = {"c": 1, "b": 2, "a": 3}
    sorted_file_path = tmp_path / "sorted.yaml"
    dump_yaml(
        data, path_to_yaml_file=sorted_file_path, sort_function=sort_alphabetically
    )
    reloaded_data = load_yaml(sorted_file_path)
    assert list(reloaded_data.keys()) == sorted(data.keys())


@pytest.mark.parametrize(
    "use_fourcipp_yaml_style, expected",
    [
        (
            False,
            """SECTION:
  - vector:
      - 1.23
      - 2
      - 3
    nested_vector:
      - - 1
        - 2.0
        - 3
      - - 4.5
        - 5
        - 2.333
  - list_with_bool:
      - 1
      - true
      - 3
    list_with_null:
      - 1
      - null
      - 4
  - list_with_string:
      - 1
      - "abc"
      - 5
    nested_list_with_string:
      - - 1
        - 2.1
      - - 2
        - "def"
""",
        ),
        (
            True,
            """SECTION:
  - vector: [1.23, 2, 3]
    nested_vector: [[1, 2.0, 3], [4.5, 5, 2.333]]
  - list_with_bool:
      - 1
      - true
      - 3
    list_with_null:
      - 1
      - null
      - 4
  - list_with_string:
      - 1
      - "abc"
      - 5
    nested_list_with_string:
      - [1, 2.1]
      - - 2
        - "def"
""",
        ),
    ],
)
def test_yaml_style(use_fourcipp_yaml_style, expected):
    """Test yaml output style."""
    data = {
        "SECTION": [
            {"vector": [1.23, 2, 3], "nested_vector": [[1, 2.0, 3], [4.5, 5, 2.333]]},
            {"list_with_bool": [1, True, 3], "list_with_null": [1, None, 4]},
            {
                "list_with_string": [1, "abc", 5],
                "nested_list_with_string": [[1, 2.1], [2, "def"]],
            },
        ]
    }
    assert (
        dict_to_yaml_string(data, use_fourcipp_yaml_style=use_fourcipp_yaml_style)
        == expected
    )



# The MIT License (MIT)
#
# Copyright (c) 2025 FourCIPP Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Test fourc input."""

import contextlib
import pathlib
import random
import subprocess
import time
from collections.abc import Callable

import pytest

from fourcipp import CONFIG
from fourcipp.fourc_input import (
    FourCInput,
    UnknownSectionException,
    sort_by_section_names,
)
from fourcipp.utils.cli import modify_input_with_defaults
from fourcipp.utils.validation import ValidationError

from ..fourcipp.legacy_io.test_element import (  # noqa: TID252
    generate_elements_from_metadatafile,
)


@pytest.fixture(name="section_names")
def fixture_section_names():
    """Section names."""
    section_name_1 = CONFIG.fourc_metadata["sections"]["specs"][0]["name"]
    section_name_2 = CONFIG.fourc_metadata["sections"]["specs"][1]["name"]
    return section_name_1, section_name_2


@pytest.fixture(name="section_names_2")
def fixture_section_names_2():
    """More section names."""
    section_name_3 = CONFIG.fourc_metadata["sections"]["specs"][2]["name"]
    section_name_4 = CONFIG.fourc_metadata["sections"]["specs"][3]["name"]
    return section_name_3, section_name_4


@pytest.fixture(name="dummy_data")
def fixture_dummy_data():
    """Dummy data for the section."""
    return {"some": "data"}


@pytest.fixture(name="fourc_input")
def fixture_fourc_input(section_names, dummy_data):
    """First input object."""
    section_name_1, section_name_2 = section_names
    fourc_input = FourCInput(
        sections={
            section_name_1: dummy_data,
            section_name_2: dummy_data,
        }
    )
    return fourc_input


@pytest.fixture(name="fourc_input_with_legacy_section")
def fixture_fourc_input_with_legacy_section(fourc_input):
    """Input object with a legacy section."""
    # Copy the input
    new_input = fourc_input.copy()

    # Add a legacy section
    new_input["DNODE-NODE TOPOLOGY"] = ["NODE 1 DNODE 1"]

    return new_input


@pytest.fixture(name="fourc_input_2")
def fixture_fourc_input_2(section_names_2, dummy_data):
    """Second input object."""
    section_name_1, section_name_2 = section_names_2
    fourc_input = FourCInput(
        sections={
            section_name_1: dummy_data,
            section_name_2: dummy_data,
        }
    )
    return fourc_input


@pytest.fixture(name="fourc_input_combined")
def fixture_joined(section_names, section_names_2, dummy_data):
    """Joined input object."""
    section_name_1, section_name_2 = section_names
    section_name_3, section_name_4 = section_names_2
    fourc_input = FourCInput(
        sections={
            section_name_1: dummy_data,
            section_name_2: dummy_data,
            section_name_3: dummy_data,
            section_name_4: dummy_data,
        }
    )
    return fourc_input


def test_str(fourc_input):
    """Assert if the string representations return a string."""
    assert isinstance(fourc_input.__str__(), str)


def test_rpr(fourc_input):
    """Assert if the string representations return a string."""
    assert isinstance(fourc_input.__repr__(), str)


def test_set_section(fourc_input, section_names, section_names_2, dummy_data):
    """Test setting section."""
    fourc_input[section_names_2[0]] = dummy_data
    fourc_input[section_names_2[1]] = dummy_data

    combined_section_names = list(section_names) + list(section_names_2)
    assert fourc_input.get_section_names() == combined_section_names


def test_set_section_failure(fourc_input, dummy_data):
    """Test setting section failure."""
    with pytest.raises(UnknownSectionException, match="Unknown section"):
        fourc_input["not existing section"] = dummy_data


def test_get_section_failure(fourc_input):
    """Test getting section failure."""
    with pytest.raises(UnknownSectionException, match="Section"):
        fourc_input["not existing section"]


def test_set_legacy_section_wrong_type(fourc_input):
    """Test setting section."""
    with pytest.raises(TypeError, match="Section"):
        fourc_input["DNODE-NODE TOPOLOGY"] = "ups not a list or dict"


@pytest.mark.parametrize("data", [{"a": "dict"}, [{"b": "also dict"}]])
def test_set_legacy_section_from_dict(fourc_input, data):
    """Test setting section from dict or list of dicts."""
    fourc_input["DNODE-NODE TOPOLOGY"] = data
    assert fourc_input["DNODE-NODE TOPOLOGY"] == data


def test_pop(fourc_input, section_names, dummy_data):
    """Test pop."""
    data = fourc_input.pop(section_names[0])
    assert data == dummy_data
    assert section_names[0] not in fourc_input


def test_pop_with_default(fourc_input, section_names_2):
    """Test pop with default."""
    data = fourc_input.pop(section_names_2[0], "default value")
    assert data == "default value"


def test_pop_but_no_default(fourc_input, section_names_2):
    """Test pop with no default."""
    with pytest.raises(UnknownSectionException, match="Section"):
        fourc_input.pop(section_names_2[0])


def test_pop_with_default_but_set(fourc_input, section_names, dummy_data):
    """Test pop with default."""
    data = fourc_input.pop(section_names[0], "default value")
    assert data == dummy_data


def test_pop_failure_unknown_section(fourc_input):
    """Test pop failure due to unknown section."""

    with pytest.raises(UnknownSectionException, match="Unknown section"):
        fourc_input.pop("invalid section", "default value")


def test_items(fourc_input, section_names, dummy_data):
    """Test itemize."""
    for k, v in fourc_input.items():
        assert k in section_names
        assert v == dummy_data


def test_contains(fourc_input):
    """Test if section is contained."""
    assert fourc_input.get_section_names()[0] in fourc_input


def test_not_contains(fourc_input):
    """Test if section is not contained."""
    assert not "some section" in fourc_input


@pytest.mark.parametrize("method", ("combine_sections", "overwrite_sections"))
def test_combine_sections_inputs(
    fourc_input, fourc_input_2, section_names, section_names_2, method
):
    """Test combine sections inputs."""
    combined_section_names = list(section_names) + list(section_names_2)

    getattr(fourc_input, method)(fourc_input_2)

    assert fourc_input.get_section_names() == combined_section_names


@pytest.mark.parametrize("method", ("combine_sections", "overwrite_sections"))
def test_combine_sections_dicts(
    fourc_input, fourc_input_2, section_names, section_names_2, method
):
    """Test combine sections from dict."""
    combined_section_names = list(section_names) + list(section_names_2)

    getattr(fourc_input, method)(fourc_input_2.sections)

    assert fourc_input.get_section_names() == combined_section_names


def test_combine_sections_failure_type(fourc_input):
    """Test combine sections inputs failure due to wrong type."""
    with pytest.raises(TypeError, match="Cannot combine sections between"):
        fourc_input.combine_sections("ups not a input object :/")


def test_combine_sections_failure_doubled_data(fourc_input):
    """Test combine sections failure due to doubled section."""
    with pytest.raises(ValueError, match="Section"):
        fourc_input.combine_sections(fourc_input.copy())


def test_combine_sections_failure_doubled_data_dict(fourc_input):
    """Test combine sections failure due to doubled section."""
    with pytest.raises(ValueError, match="Section"):
        fourc_input.combine_sections(fourc_input.copy().sections)


def test_overwrite_sections_new_section_dict(fourc_input):
    """Test overwriting sections that does not exist."""
    fourc_input.overwrite_sections({"TITLE": "new title"})
    assert fourc_input["TITLE"] == "new title"


def test_overwrite_sections_existing_section_dict(fourc_input):
    """Test overwriting sections that already exist."""
    fourc_input["TITLE"] = "new title"
    fourc_input.overwrite_sections({"TITLE": "super new title"})
    assert fourc_input["TITLE"] == "super new title"


def test_overwrite_sections_new_section_input(fourc_input):
    """Test overwriting sections that does not exist."""
    fourc_input.overwrite_sections(FourCInput({"TITLE": "new title"}))
    assert fourc_input["TITLE"] == "new title"


def test_overwrite_sections_existing_section_input(fourc_input):
    """Test overwriting sections that already exist."""
    fourc_input["TITLE"] = "new title"
    fourc_input.overwrite_sections(FourCInput({"TITLE": "super new title"}))
    assert fourc_input["TITLE"] == "super new title"


def test_apply_default(tmp_path):
    """Test using a default file.

    4C default file with a section that is not in the input file and one
    that is. A section containing a list should not be added.
    """

    # Creating the default file
    path_to_default = tmp_path / "default.4C.yaml"
    default_input = FourCInput({"PROBLEM TYPE": {"PROBLEMTYPE": "default problemtype"}})
    default_input["PROBLEM SIZE"] = {"DIM": 3}
    default_input.dump(path_to_default)

    # writing some entries to the current fourc_input
    fourc_input = FourCInput({"SOLVER 1": {"NAME": "current title"}})
    fourc_input["PROBLEM TYPE"] = {"PROBLEMTYPE": "current problemtype"}

    # Applying the defaults
    fourc_input.apply_user_defaults(path_to_default)

    # Only in current input
    assert fourc_input["SOLVER 1"] == {"NAME": "current title"}
    # Defined in both default and current input -> do not overwrite with the default
    assert fourc_input["PROBLEM TYPE"] == {"PROBLEMTYPE": "current problemtype"}
    # Defined in default, not in current input -> should be taken
    assert fourc_input["PROBLEM SIZE"] == {"DIM": 3}
    # should not be taken even though it is in the default file, because it is a list
    assert not "MATERIALS" in fourc_input


def test_apply_user_defaults(
    fourc_input, fourc_input_2, fourc_input_combined, tmp_path
):
    """Test for applying user defaults from the file assigned in the config."""

    # change user_default_path to tmp_path/user_defaults.4C.yaml
    # This cannot be done using the cli function change_user_defaults_path
    CONFIG.user_defaults_path = tmp_path / "user_defaults.4C.yaml"
    # dump fourc_input to tmp_path/current_input.4C.yaml
    fourc_input.dump(tmp_path / "current_input.4C.yaml")
    # dump fourc_input_2 to user_default_path
    fourc_input_2.dump(tmp_path / "user_defaults.4C.yaml")
    # Now comes the function to be tested:
    # apply_user_defaults to fourc_input
    modify_input_with_defaults(tmp_path / "current_input.4C.yaml", True)
    # should result in fourc_input_combined
    defaulted_input = FourCInput.from_4C_yaml(tmp_path / "current_input.4C.yaml")
    assert defaulted_input == fourc_input_combined


def test_add(fourc_input, fourc_input_2, fourc_input_combined):
    """Test adding inputs."""
    added_input = fourc_input + fourc_input_2

    assert added_input == fourc_input_combined
    assert added_input != fourc_input
    assert added_input != fourc_input_2


def test_equal(fourc_input):
    """Test for equal inputs."""
    assert fourc_input.sections == fourc_input.copy().sections


def test_equal_failure(fourc_input):
    """Test for equal failure."""
    with pytest.raises(TypeError, match="Can not compare types"):
        assert fourc_input == "wrong type"


def test_not_equal(fourc_input, fourc_input_2):
    """Test for non-equal inputs."""
    assert not fourc_input == fourc_input_2


def test_load_includes(fourc_input, fourc_input_2, fourc_input_combined, tmp_path):
    """Test loading includes."""
    path_to_other_sections = tmp_path / "split_data.4C.yaml"
    fourc_input_2.dump(path_to_other_sections)

    fourc_input["INCLUDES"] = [str(path_to_other_sections)]

    fourc_input.load_includes()

    assert fourc_input == fourc_input_combined


def test_split(fourc_input, fourc_input_2, fourc_input_combined, section_names_2):
    """Test split."""
    first, second = fourc_input_combined.split(section_names_2)

    assert first == fourc_input
    assert second == fourc_input_2


def test_dump_with_includes(
    fourc_input_combined, tmp_path, section_names, section_names_2
):
    """Test dump with includes."""
    path_1 = tmp_path / "path_1.4C.yaml"
    path_2 = tmp_path / "path_2.4C.yaml"

    fourc_input_combined.dump_with_includes(section_names, path_1, path_2)

    reloaded = FourCInput.from_4C_yaml(path_1)
    assert reloaded.get_section_names() == list(section_names_2) + ["INCLUDES"]


def test_dump_with_includes_invert_sections(
    fourc_input_combined, tmp_path, section_names
):
    """Test dump with includes with invert sections."""
    path_1 = tmp_path / "path_1.4C.yaml"
    path_2 = tmp_path / "path_2.4C.yaml"

    fourc_input_combined.dump_with_includes(
        section_names, path_1, path_2, invert_sections=True
    )

    reloaded = FourCInput.from_4C_yaml(path_1)
    assert reloaded.get_section_names() == list(section_names) + ["INCLUDES"]


def get_4C_test_input_files():
    """Get all input test files in 4C docker image."""
    test_files_directory = pathlib.Path("/home/user/4C/tests/input_files")

    if not test_files_directory.exists():
        return []

    files = [
        "ale2d_solid_lin.4C.yaml",
        "beam3eb_genalpha_lineload_dynamic.4C.yaml",
        "beam3r_herm2line3_static_beam_to_solid_volume_meshtying_2d-3d.4C.yaml",
        "contact3D_quad_tet10.4C.yaml",
        "elch_2D_porousMediumHomo_SSPP.4C.yaml",
        "f2_nurbs9_dc_drt.4C.yaml",
        "fsi_dc_mono_slss_msht.4C.yaml",
        "particle_dem_1d_adhesion_RegDMT.4C.yaml",
        "poro_3D_tet4.4C.yaml",
        "reduced_lung_3_aw_2_tu.4C.yaml",
        "scatra_chemo_h27.4C.yaml",
        "solid_ele_tet4_Standard_linear.4C.yaml",
        "tsi_meshtying_nurbs.4C.yaml",
        "xfsi_comp_struct_fsi_2D_mono_slip.4C.yaml",
        "xfsi_3D_boxes.4C.yaml",  # test domain
    ]

    return [str(test_files_directory / file) for file in files]


FOURC_TEST_INPUT_FILES = get_4C_test_input_files()


class SubprocessError(Exception):
    """Subprocess failure."""


@pytest.mark.skipif(CONFIG.name != "4C_docker_main", reason="Not using docker config.")
@pytest.mark.parametrize("fourc_file", FOURC_TEST_INPUT_FILES)
def test_roundtrip_test(fourc_file, tmp_path):
    """Roundtrip test."""
    fourc_file = pathlib.Path(fourc_file)

    # Load 4C input test file
    fourc_input = FourCInput.from_4C_yaml(fourc_file)

    # Delete the original file (we need to dump the file in the original location to ensure that associated files, i.e., xml solver files, are still present)
    fourc_file.unlink()

    # Dump out again
    fourc_input.dump(fourc_file, validate=True)

    # Command
    command = (
        f"/home/user/4C/build/4C {fourc_file} xxx > {tmp_path / 'output.log'} 2>&1"
    )

    # Run 4C with the dumped input
    return_code = subprocess.call(command, shell=True)  # nosec

    # Exit code -> 4C failed
    if return_code:
        raise SubprocessError(
            f"Input file failed for {fourc_file}.\n\n4C command: {command}\n\nOutput: {(tmp_path / 'output.log').read_text()}"
        )


@pytest.mark.skipif(CONFIG.name != "4C_docker_main", reason="Not using docker config.")
@pytest.mark.parametrize(
    "fourc_file",
    [
        str(f.resolve())
        for f in pathlib.Path("/home/user/4C/tests/input_files").glob("*.4C.yaml")
    ],
)
def test_readin_all_test_files(fourc_file):
    """Read all known 4C files and check if valid."""
    fourc_file = pathlib.Path(fourc_file)

    # Load 4C input test file
    fourc_input = FourCInput.from_4C_yaml(fourc_file)

    # Check if it is valid
    fourc_input.validate()


def test_extract_header_sections(fourc_input, fourc_input_with_legacy_section):
    """Test the header extraction."""

    # Extract the header
    header = fourc_input_with_legacy_section.extract_header()

    assert header == fourc_input


def test_load_from_yaml(fourc_input_with_legacy_section, tmp_path):
    """Test load from yaml file."""
    path_to_yaml = tmp_path / "fourc_input.4C.yaml"
    fourc_input_with_legacy_section.dump(path_to_yaml)

    assert fourc_input_with_legacy_section == FourCInput.from_4C_yaml(path_to_yaml)


def test_load_from_yaml_header_only(
    fourc_input, fourc_input_with_legacy_section, tmp_path
):
    """Test load from yaml file using header only."""
    path_to_yaml = tmp_path / "fourc_input.4C.yaml"
    fourc_input_with_legacy_section.dump(path_to_yaml)

    assert fourc_input == FourCInput.from_4C_yaml(path_to_yaml, header_only=True)


def test_compare(fourc_input):
    """Test compare function."""
    copy_input = fourc_input.copy()
    assert fourc_input.compare(copy_input)


def test_compare_failure(fourc_input, fourc_input_2):
    """Test compare function failure."""
    assert not fourc_input.compare(fourc_input_2)


def test_compare_failure_with_exception(fourc_input, fourc_input_2):
    """Test compare function failure."""
    with pytest.raises(AssertionError):
        fourc_input.compare(fourc_input_2, raise_exception=True)


@pytest.mark.parametrize(
    "fourc_input,error_context, sections_only",
    [
        (
            FourCInput(sections={"TITLE": "some title"}),
            pytest.raises(ValidationError),
            False,
        ),
        (
            FourCInput(
                sections={
                    "TITLE": "some title",
                    "PROBLEM TYPE": {"PROBLEMTYPE": "Fluid"},
                }
            ),
            contextlib.nullcontext(),  # No error
            False,
        ),
        (
            FourCInput(sections={"TITLE": "some title"}),
            contextlib.nullcontext(),
            True,
        ),
        (
            FourCInput(
                sections={
                    "TITLE": "some title",
                    "PROBLEM TYPE": {"PROBLEMTYPE": "Fluid"},
                }
            ),
            contextlib.nullcontext(),  # No error
            True,
        ),
    ],
)
def test_validation(fourc_input, error_context, sections_only):
    """Test the validation."""
    with error_context:
        fourc_input.validate(sections_only=sections_only)


def test_sort_by_section_names():
    """Test sorting by section names."""

    # create list of typed sections without title and required sections
    typed_sections = [
        sec
        for sec in CONFIG.sections.typed_sections
        if sec != CONFIG.fourc_metadata["metadata"]["description_section_name"]
        and sec not in set(CONFIG.fourc_json_schema["required"])
    ]

    # also use end subset to also add some lowercase sections
    typed_sections = typed_sections[:15] + typed_sections[-15:]
    typed_sections = sorted(typed_sections, key=str.lower)

    # use first 5 'DESIGN * ' sections
    design_sections = [
        s for s in CONFIG.sections.typed_sections if s.startswith("DESIGN")
    ][:5]
    design_sections = sorted(design_sections, key=str.lower)

    # create some FUNCT sections
    function_sections = [f"FUNCT{i}" for i in [1, 2, 9, 10, 33]]

    correct_section_order = (
        [CONFIG.fourc_metadata["metadata"]["description_section_name"]]
        + CONFIG.fourc_json_schema["required"]
        + typed_sections
        + ["MATERIALS"]
        + design_sections
        + function_sections
        + CONFIG.sections.legacy_sections
    )

    shuffled_section_order = correct_section_order.copy()
    random.seed(42)
    random.shuffle(shuffled_section_order)

    shuffled_data = {k: 1 for k in shuffled_section_order}

    sorted_data = sort_by_section_names(shuffled_data)

    assert list(sorted_data.keys()) == correct_section_order


## performance tests


def create_dummy_elements() -> dict:
    """Create dummy elements for the performance test.

    Loops over all elements from the metadata file and creates
    3000 dummy elements for each element type.

    Returns:
        dict: Dictionary with dummy elements.
    """

    reference_elements = generate_elements_from_metadatafile()

    dummy_elements: dict[str, list[str]] = {"STRUCTURE ELEMENTS": []}

    for element in reference_elements:
        for i in range(1, 3000):
            dummy_elements["STRUCTURE ELEMENTS"].append(
                f"{i} {' '.join(element.split()[1:])}"
            )

    return dummy_elements


def evaluate_execution_time(fct: Callable, args: dict) -> float:
    """Evaluate execution time of a function.

    Args:
        fct: Function to test.
        args: Arguments to pass to the function.

    Returns:
        float: Execution time in seconds.
    """

    start_time = time.time()
    fct(**args)
    end_time = time.time()

    return end_time - start_time


def save_timings(timings: dict, file: str) -> None:
    """Save timings to a file.

    Args:
        timings: Dictionary with timings.
        file: Name of the markdown file to save the timings.
    """

    if not file.endswith(".md"):
        raise ValueError("File must be a markdown file ending with .md!")

    with open(file, "w", encoding="utf-8") as f:
        f.write("# Performance Timings :rocket:\n\n")
        f.write("| Operation         | Time (seconds)  |\n")
        f.write("|-------------------|-----------------|\n")
        for operation, time_taken in timings.items():
            f.write(f"| {operation:<17} | {time_taken:10.6f}      |\n")


@pytest.mark.performance
def test_performance(tmp_path) -> None:
    """Test performance of core functions of FourCInput."""

    dummy_elements = create_dummy_elements()

    fourc_input = FourCInput()

    timings = {}

    # Evaluate execution time of adding elements to input file
    timings["add_elements"] = evaluate_execution_time(
        fourc_input.combine_sections, args={"other": dummy_elements}
    )

    # Evaluate execution time of validating the input file
    timings["validate"] = evaluate_execution_time(
        fourc_input.validate, args={"sections_only": True}
    )

    # Evaluate performance of dumping the input file
    timings["dump"] = evaluate_execution_time(
        fourc_input.dump,
        args={"input_file_path": tmp_path / "performance_test.4C.yaml"},
    )

    # Evaluate performance of loading the input file
    timings["load_from_file"] = evaluate_execution_time(
        FourCInput.from_4C_yaml,
        args={"input_file_path": tmp_path / "performance_test.4C.yaml"},
    )

    save_timings(timings, "timings.md")



# The MIT License (MIT)
#
# Copyright (c) 2025 FourCIPP Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Not set tests."""

import pytest

from fourcipp.utils.not_set import NOT_SET, NotSet, check_if_set, pop_arguments


@pytest.mark.parametrize(
    "obj, expected_result", [("a", True), (2, True), ({2: 3}, True), (NOT_SET, False)]
)
def test_check_if_set(obj, expected_result):
    """Test check if set."""
    assert check_if_set(obj) == expected_result


def test_pop_arguments_without_default():
    """Assert pop arguments without default."""
    assert pop_arguments("key") == ("key",)


def test_pop_arguments_with_default():
    """Assert pop arguments with default."""
    assert pop_arguments("key", "default") == ("key", "default")


def test_expected():
    """Test expected."""
    a = NotSet(int)
    assert a.expected == int



# The MIT License (MIT)
#
# Copyright (c) 2025 FourCIPP Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Test inline dat utils."""

from functools import partial

import pytest

from fourcipp.legacy_io.inline_dat import (
    _entry_casting_factory,
    _extract_entry,
    _extract_enum,
    casting_factory,
    inline_dat_read,
)


def test_extract_enum_invalid_choice():
    """Test enum extract for invalid choice."""
    with pytest.raises(ValueError, match="Unknown entry"):
        _extract_enum(["a"], choices=["b", "c"])


def test_entry_casting_factory_invalid_type():
    """Test reading invalid types."""
    with pytest.raises(NotImplementedError, match="Entry type"):
        _entry_casting_factory({"type": "invalid"})


@pytest.mark.parametrize(
    "value,expected_value",
    [
        ("true", True),
        ("yes", True),
        ("1", True),
        ("false", False),
        ("no", False),
        ("0", False),
    ],
)
def test_entry_casting_factory_bool_cast(value, expected_value):
    """Test bool casting."""
    spec = {"type": "bool"}
    cast_func = _entry_casting_factory(spec)

    assert cast_func([value]) is expected_value


@pytest.mark.parametrize(
    "invalid_input", ["True", "False", "some text", "TRUE", "FALSE"]
)
def test_entry_casting_factory_invalid_bool_cast(invalid_input):
    """Test invalid bool casting."""
    spec = {"type": "bool"}
    cast_func = _entry_casting_factory(spec)

    with pytest.raises(KeyError):
        cast_func([invalid_input])


def test_casting_factory_invalid_type():
    """Test reading invalid types."""
    with pytest.raises(NotImplementedError, match="First entry has to be an all_of"):
        casting_factory({"type": "invalid"})


def test_doubled_inline_keyword():
    """Test if parameter is set twice."""
    with pytest.raises(KeyError, match="The entry a was provided already"):
        inline_dat_read(
            ["a", "1", "b", "1", "a", "2"],
            keyword_casting={
                "a": partial(_extract_entry, extractor=int),
                "b": partial(_extract_entry, extractor=int),
            },
        )



# The MIT License (MIT)
#
# Copyright (c) 2025 FourCIPP Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Configuration utils."""

from __future__ import annotations

import copy
import pathlib
from dataclasses import dataclass, field

from loguru import logger

from fourcipp.utils.type_hinting import Path, T
from fourcipp.utils.yaml_io import dump_yaml, load_yaml

CONFIG_PACKAGE: pathlib.Path = pathlib.Path(__file__).parents[1] / "config"
CONFIG_FILE: pathlib.Path = CONFIG_PACKAGE / "config.yaml"


class Sections:
    def __init__(self, legacy_sections: list[str], typed_sections: list[str]):
        """Sections data container.

        Args:
            legacy_sections: Legacy sections, i.e., their information is not provided in the schema file
            typed_sections: Typed sections, non-legacy sections natively supported by the schema
        """
        self.legacy_sections: list[str] = legacy_sections
        self.typed_sections: list[str] = typed_sections
        self.all_sections: list[str] = typed_sections + legacy_sections

    @classmethod
    def from_metadata(cls, fourc_metadata: dict) -> Sections:
        """Get section names from metadata.

        Args:
            fourc_metadata (dict): 4C metadata

        Returns:
            Sections: sections object
        """
        description_section = fourc_metadata["metadata"]["description_section_name"]
        sections = [description_section] + [
            section["name"] for section in fourc_metadata["sections"]["specs"]
        ]

        legacy_sections = list(fourc_metadata["legacy_string_sections"])

        return cls(legacy_sections, sections)


@dataclass
class ConfigProfile:
    """Fourcipp configuration profile.

    Attributes:
        name: Name of the configuration profile
        description: Description of the profile
        fourc_metadata_path: Path to metadata yaml file
        json_schema_path: Path to json schema path
        user_defaults_path: Path to user specific defaults
    """

    name: str
    description: str
    fourc_metadata_path: Path
    fourc_json_schema_path: Path
    user_defaults_path: Path | None = None
    fourc_metadata: dict = field(init=False)
    fourc_json_schema: dict = field(init=False)
    sections: Sections = field(init=False)

    def __post_init__(self) -> None:
        """Update stuff."""
        self.fourc_metadata_path = pathlib.Path(self.fourc_metadata_path)
        self.fourc_metadata = ConfigProfile._resolve_references(
            ConfigProfile._load_data_from_path(self.fourc_metadata_path)
        )
        self.sections = Sections.from_metadata(self.fourc_metadata)

        self.fourc_json_schema_path = pathlib.Path(self.fourc_json_schema_path)
        self.fourc_json_schema = ConfigProfile._load_data_from_path(
            self.fourc_json_schema_path
        )

        if self.user_defaults_path is not None:
            self.user_defaults_path = pathlib.Path(self.user_defaults_path)
            if not self.user_defaults_path.is_file():
                raise FileNotFoundError(
                    f"User defaults file '{self.user_defaults_path}' does not exist."
                )

    @staticmethod
    def _load_data_from_path(path: Path) -> dict:
        """Load data from path."""
        if not pathlib.Path(path).is_absolute():
            # Assumption: Path is relative to FourCIPP config package
            logger.debug(
                f"Path {path} is a relative path. The absolute path is set to {CONFIG_PACKAGE / path}"
            )
            path = CONFIG_PACKAGE / path
        return load_yaml(path)

    @staticmethod
    def _resolve_references(metadata_dict: dict) -> dict:
        """Resolve references in the 4C metadata file.

        Args:
            metadata_dict: 4C metadata

        Returns:
            metadata_dict without references
        """

        references = metadata_dict.pop("$references")

        def insert_references(metadata: T, references: dict) -> T:
            """Iterate nested dict and insert references.

            Args:
                metadata: Metadata to check for references
                references: Dict with all the references

            Returns:
                metadata with resolved references
            """
            if isinstance(metadata, dict):
                if "$ref" in metadata:
                    # Add an actual copy of the data
                    metadata = copy.deepcopy(references[metadata.pop("$ref")])
                else:
                    for k in metadata:
                        metadata[k] = insert_references(metadata[k], references)
            elif isinstance(metadata, list):
                for i, e in enumerate(metadata):
                    metadata[i] = insert_references(e, references)

            return metadata

        metadata_dict = insert_references(metadata_dict, references)

        return metadata_dict

    def __str__(self) -> str:
        """String method for the config."""

        def add_keyword(name: str, data: object) -> str:
            """Create keyword description line."""
            return f"\n - {name}: {data}"

        s = f"FourCIPP configuration '{self.name}'"
        s += add_keyword("Configuration file", CONFIG_FILE)
        s += add_keyword("Description", self.description)
        s += add_keyword("4C metadata path", self.fourc_json_schema_path)
        s += add_keyword("4C JSON schema path", self.fourc_json_schema_path)
        s += add_keyword("User default path", self.user_defaults_path)

        return s


def load_config() -> ConfigProfile:
    """Set config profile.

    Args:
        profile: Config profile to be set.

    Returns:
        user config.
    """
    config_data: dict = load_yaml(CONFIG_FILE)
    profile_name = config_data["profile"]
    profile = config_data["profiles"][profile_name]
    logger.debug(f"Reading config profile {profile}")

    config = ConfigProfile(name=profile_name, **profile)
    logger.debug(config)
    return config


def change_profile(profile: str) -> None:
    """Change config profile.

    Args:
        profile: Profil name to set
    """
    config_data: dict = load_yaml(CONFIG_FILE)

    if profile not in config_data["profiles"]:
        known_profiles = ", ".join(config_data["profiles"])
        raise KeyError(
            f"Profile {profile} unknown. Known profiles are: {known_profiles}"
        )
    config_data["profile"] = profile
    logger.info(f"Changing to config profile '{profile}'")
    dump_yaml(config_data, CONFIG_FILE)


def show_config() -> None:
    """Show FourCIPP config."""
    logger.info("Fourcipp configuration")
    logger.info(f"  Config file: {CONFIG_FILE.resolve()}")
    logger.info("  Contents:")
    logger.info("    " + "\n    ".join(CONFIG_FILE.read_text().split("\n")))



# The MIT License (MIT)
#
# Copyright (c) 2025 FourCIPP Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Utils for legacy io testing."""

from fourcipp.legacy_io.element import to_dat_string

# This map is used to generate elements from the metadata file
_TESTING_MAP = {"int": 4, "double": 2.11, "string": "text", "bool": True}


def iterate_all_of(all_of):
    """Iterate over all_of and one_ofs.

    Args:
        all_of (dict): all_of dictionary

    Yields:
        dict: specs of the all_of
    """
    for spec in all_of["specs"]:
        if spec["type"] == "all_of":
            yield from iterate_all_of(spec)
        elif spec["type"] == "one_of":
            for spec in spec["specs"]:
                yield from iterate_all_of(spec)
        else:
            yield spec
    return


def reference_value_from_all_of(all_of):
    """Generate line from all_of.

    Args:
        all_of (dict): Metadata dict for the group

    Returns:
        str: inline
    """

    keyword_parameters = {}
    positional_parameters = {}

    for parameter in iterate_all_of(all_of):
        entry = None
        if parameter["type"] in _TESTING_MAP:
            entry = _TESTING_MAP[parameter["type"]]
        elif parameter["type"] == "vector":
            entry = [_TESTING_MAP[parameter["value_type"]["type"]]] * parameter["size"]
        elif parameter["type"] == "enum":
            entry = parameter["choices"][0]["name"]
        else:
            raise ValueError(f"Could not create testing line from {parameter['type']}")

        if parameter["name"].startswith("_positional"):
            _, _, position, _ = parameter["name"].split("_", 3)
            positional_parameters[int(position)] = entry
        else:
            keyword_parameters[parameter["name"]] = to_dat_string(entry)

    line_list = [
        to_dat_string(positional_parameters[e])
        for e in sorted(positional_parameters.keys())
    ]
    for e in keyword_parameters.items():
        line_list.extend(e)
    return "  ".join(line_list)
